import sys
from hcache import cached
import scipy as sp
import numpy as np
from .covar_base import Covariance
import scipy.spatial as SS
from . import sqexp


# TODO would be cleaner to inherit from squared exponential ?
class ZKZCov(Covariance):
    """
    squared exponential covariance function
    """
    def __init__(self,X, Kin, remove_diag=True):
        """
        X   dim x d input matrix
        """
        super(ZKZCov, self).__init__()
        self._X = None
        self.se = sqexp.SQExpCov(X)
        self.X = X
        self.Kin = Kin
        self.rm_diag = remove_diag
        # self._initParams()

    def get_input_dim(self):
        return self.X.shape[1]

    #####################
    # Properties
    #####################
    @property
    def scale(self):
        return self.se.scale

    @property
    def length(self):
        return self.se.length

    # TODO not true in the general case -> to change
    @property
    def dim(self):
        return self.se.dim

    @property
    def scale_ste(self):
        # if self.getFIinv() is None:
        #     R = None
        # else:
        #     R = sp.sqrt(self.getFIinv()[0,0])
        # return R
        return self.se.scale_ste

    @property
    def length_ste(self):
        # if self.getFIinv() is None:
        #     R = None
        # else:
        #     R = sp.sqrt(self.getFIinv()[1,1])
        # return R
        return self.se.length_ste

    @property
    def X(self):
        return self._X

    #####################
    # Setters
    #####################
    @scale.setter
    def scale(self, value):
        self.se.scale = value
        self.clear_all()
        self._notify()

    @length.setter
    def length(self, value):
        self.se.length = value
        self.clear_all()
        self._notify()

    @X.setter
    def X(self,value):
        self.se.X = value

    @Covariance.use_to_predict.setter
    def use_to_predict(self, value):
        assert self.Xstar is not None, 'set Xstar!'
        self._use_to_predict = value
        self._notify()

    #####################
    # Params handling
    #####################
    def getNumberParams(self):
        return self.se.getNumberParams()

    def setParams(self, params):
        self.se.setParams(params)
        self.clear_all()

    def getParams(self):
        return self.se.getParams()

    def _calcNumberParams(self):
        self.n_params = 2

    #####################
    # Activation handling
    #####################

    @property
    def act_scale(self):
        return self.se._scale_act

    @act_scale.setter
    def act_scale(self, act):
        self.se._scale_act = bool(act)
        self._notify()

    @property
    def act_length(self):
        return self.se._length_act

    @act_length.setter
    def act_length(self, act):
        self.se._length_act = bool(act)
        self._notify()

    #####################
    # Cached
    #####################
    @cached('covar_base')
    def K(self):
        z = self.se.K()
        if self.rm_diag:
            z -= np.eye(z.shape[0]) * z.diagonal()
        tmp = z.dot(self.Kin.dot(z.transpose()))
        return tmp

    @cached('covar_base')
    def K_grad_i(self, i):
        grad_tmp = self.se.K_grad_i(i)
        se_K_tmp = self.se.K()
        if self.rm_diag:
            grad_tmp -= grad_tmp.diagonal() * np.eye(grad_tmp.shape[0])
            se_K_tmp -= se_K_tmp.diagonal() * np.eye(se_K_tmp.shape[0])
        r = grad_tmp.dot(self.Kin.dot(se_K_tmp.transpose()))
        r += se_K_tmp.dot(self.Kin.dot(grad_tmp.transpose()))
        return r

    ####################
    # Interpretable Params
    ####################
    def getInterParams(self):
        return sp.array([self.se.scale, self.se.length])

    # def K_grad_interParam_i(self,i):
    #     if i==0:
    #         r = sp.exp(-self.E()/(2*self.se.length))
    #     else:
    #         A = sp.exp(-self.E()/(2*self.se.length))*self.se.E()
    #         r = self.se.scale * A / (2*self.se.length**2)
    #     return r



import pdb
import scipy as SP
import scipy.linalg as LA
import copy
import sys
from limix.core.linalg.linalg_matrix import jitChol
import limix.core.likelihood.likelihood_base

import logging as LG


class GP(object):
    """
    Gaussian Process regression class. Holds all information for the GP regression to take place.
    """

    def __init__(self,Y,K):
        """
        y:        phenotype vector
        K:        sample-to-sample covariance matrix (includes noise!)
        """
        self.covar = K
        self.setY(Y)

    def setY(self,Y):
        """
        set pheno
        """
        if Y.ndim==1: Y = Y[:,SP.newaxis]
        self.n,self.t = Y.shape
        self.Y = Y
        self.Y_has_changed = True

    def getParams(self):
        params = {}
        params['covar'] = self.covar.getParams()
        return params

    def setParams(self,params):
        self.params = params
        self.updateParams()

    def updateParams(self):
        self.covar.setParams(self.params['covar'])

    def LML(self,params=None):
        """
        evalutes the log marginal likelihood for the given hyperparameters

        hyperparams
        """
        if params is not None:
            self.setParams(params)

        KV = self._update_cache()
        alpha = KV['alpha']
        L = KV['L']

        lml_quad = 0.5 * (alpha*self.Y).sum()
        lml_det = self.t *SP.log(SP.diag(L)).sum()
        lml_const = 0.5*self.n*self.t*SP.log(2*SP.pi)
        LML = lml_quad + lml_det + lml_const
        return LML


    def LMLgrad(self,params=None):
        """
        evaluates the gradient of the log marginal likelihood for the given hyperparameters
        """
        if params is not None:
            self.setParams(params)
        KV = self._update_cache()
        W = KV['W']
        LMLgrad = SP.zeros(self.covar.n_params)
        for i in range(self.covar.n_params):
            Kd = self.covar.Kgrad_param(i)
            LMLgrad[i] = 0.5 * (W*Kd).sum()
        return {'covar':LMLgrad}


    def predict(self,Xstar):
        """
        predict on Xstar
        """
        KV = self._update_cache()
        self.covar.setXstar(Xstar)
        Kstar = self.covar.Kcross()
        Ystar = SP.dot(Kstar,KV['alpha'])
        return Ystar

    def _update_cache(self):
        """
        INPUT:
        hyperparams:  dictionary
        OUTPUT: dictionary with the fields
        K:     kernel
        Kinv:  inverse of the kernel
        L:     chol(K)
        alpha: solve(K,y)
        W:     D*Kinv * alpha*alpha^T
        """
        cov_params_have_changed = self.covar.params_have_changed

        if cov_params_have_changed or self.Y_has_changed:
            K = self.covar.K()
            L = LA.cholesky(K).T# lower triangular
            Kinv = LA.cho_solve((L,True),SP.eye(L.shape[0]))
            alpha = LA.cho_solve((L,True),self.Y)
            W = self.t*Kinv - SP.dot(alpha,alpha.T)
            self._covar_cache = {}
            self._covar_cache['K'] = K
            self._covar_cache['Kinv'] = Kinv
            self._covar_cache['L'] = L
            self._covar_cache['alpha'] = alpha
            self._covar_cache['W'] = W

        return self._covar_cache


    def checkGradient(self,h=1e-6,verbose=True):
        """ utility function to check the gradient of the gp """
        grad_an = self.LMLgrad()
        grad_num = {}
        params0 = self.params.copy()
        for key in list(self.params.keys()):
            paramsL = params0.copy()
            paramsR = params0.copy()
            grad_num[key] = SP.zeros_like(self.params[key])
            e = SP.zeros(self.params[key].shape[0])
            for i in range(self.params[key].shape[0]):
                e[i] = 1
                paramsL[key]=params0[key]-h*e
                paramsR[key]=params0[key]+h*e
                lml_L = self.LML(paramsL)
                lml_R = self.LML(paramsR)
                grad_num[key][i] = (lml_R-lml_L)/(2*h)
                e[i] = 0
            if verbose:
                print(('%s:'%key))
                print((abs(grad_an[key]-grad_num[key])))
                print('')
        self.setParams(params0)



import sys
import scipy as SP
import h5py
import pdb

def vec(M):
    return M.reshape((M.size, 1), order = 'F')

def to_list(x):
    if x is None:               r = []
    elif type(x) is not list:   r = [x]
    else:                       r = x
    return r

def smartAppend(table,name,value):
    """
    helper function for apppending in a dictionary  
    """ 
    if name not in list(table.keys()):
        table[name] = []
    table[name].append(value)

def smartSum(x,key,value):
    """ create a new page in x if key is not a page of x
        otherwise add value to x[key] """
    if key not in list(x.keys()):
        x[key] = value
    else:   x[key]+=value

def dumpDictHdf5(RV,o):
    """ Dump a dictionary where each page is a list or an array """
    for key in list(RV.keys()):
        o.create_dataset(name=key,data=SP.array(RV[key]),chunks=True,compression='gzip')

def smartDumpDictHdf5(RV,o):
    """ Dump a dictionary where each page is a list or an array or still a dictionary (in this case, it iterates)"""
    for key in list(RV.keys()):
        if type(RV[key])==dict:
            g = o.create_group(key)
            smartDumpDictHdf5(RV[key],g)
        else:
            o.create_dataset(name=key,data=SP.array(RV[key]),chunks=True,compression='gzip')




import sys
from limix.core.mean import MeanKronSum
from limix.core.covar import Cov3KronSumLR
from hcache import Cached, cached
from limix.utils.util_functions import vec

import pdb
import numpy as NP
import scipy as sp
import scipy.linalg as la
import time as TIME
from .gp_base import GP
from limix.core.gp import GP2KronSum
from limix.core.covar.cov_reml import cov_reml

class GP3KronSumLR(GP2KronSum):
    """
    Gaussian Process with a 3kronSumLR Covariance and a mean that is a sum of Kronecker products:
        vec(Y) ~ N( vec( \sum_i F_i B_i A_i), Cr \kron GG.T + Cg \kron R + Cn \kron I )
    The current implementation of this class does not support fixed effects!
    Notation:
        N = number of samples
        P = number of traits
        Y = [N, P] phenotype matrix
        F_i = sample fixed effect design for term i
        A_i = trait fixed effect design for term i
        B_i = effect sizes of fixed effect term i
        Cr = column covariance matrix for low-rank term 
        Cg = column covariance matrix for full rank non-noise term
        Cn = column covariance matrix for noise term
        R = row covariance matrix for non-noise term
        rank_c = rank of low-rank col covariance
        rank_r = rank of low-rank row covariance
    """

    def __init__(self,Y = None, Cg = None, Cn = None, R = None, S_R = None, U_R = None, G = None, Cr=None, rank = 1):
        """
        Args:
            Y:      [N, P] phenotype matrix
            Cg:     Limix covariance matrix for Cg (dimension P)
            Cn:     Limix covariance matrix for Cn (dimension P)
            G:      [N, rank_r] numpy covariance matrix for G
            R:      [N, N] numpy semidemidefinite covariance matrix for R.
                    In alternative to R, S_R and U_R can be specified.
            S_R:    N vector of eigenvalues of R
            U_R:    [N, N] eigenvector matrix of R
            rank:   rank of column low-rank covariance (default = 1)
        """
        covar = Cov3KronSumLR(Cg=Cg, Cn=Cn, R=R, G=G, Cr=Cr, rank=rank, S_R=S_R, U_R=U_R)
        mean  = MeanKronSum(Y = Y)
        GP.__init__(self, covar = covar, mean = mean)

    def _observe(self):
        self.covar.register(self.row_cov_has_changed, 'row_cov')
        self.covar.register(self.col_cov_has_changed, 'col_cov')
        self.covar.register(self.G_has_changed, 'G')
        self.mean.register(self.pheno_has_changed, 'pheno')
        self.mean.register(self.designs_have_changed, 'designs')

    def G_has_changed(self):
        self.clear_cache('G')
        self.clear_all()

    ######################
    # Transformed phenotype
    ######################
    @cached(['pheno','col_cov', 'row_cov', 'G'])
    def ve_WrDLrYLcWc(self):
        return vec(sp.dot(self.covar.Wr().T, sp.dot(self.DLrYLc(), self.covar.Wc())))

    @cached(['pheno','col_cov', 'row_cov', 'G'])
    def Hi_ve_WrDLrYLcWc(self):
        return la.cho_solve((self.covar.H_chol(), True), self.ve_WrDLrYLcWc())

    @cached(['pheno','col_cov', 'row_cov', 'G'])
    def vei_HiveWrDLrYLcWc(self):
        return self.Hi_ve_WrDLrYLcWc().reshape((self.covar.rank_r, self.covar.rank_c), order = 'F')

    @cached(['pheno','col_cov','row_cov','G'])
    def DLrYLc_m_DWr_veiHiveWrDLrYLcWc_Wc(self):
        R  = sp.dot(self.covar.Wr(), sp.dot(self.vei_HiveWrDLrYLcWc(), self.covar.Wc().T))
        R *= -self.covar.D()
        R += self.DLrYLc()
        return R

    @cached(['pheno','col_cov','row_cov','G'])
    def WrWr_DLrYLc_m_DWr_veiHiveWrDLrYLcWc_Wc(self):
        return sp.dot(self.covar.Wr(), sp.dot(self.covar.Wr().T, self.DLrYLc_m_DWr_veiHiveWrDLrYLcWc_Wc()))

    @cached(['pheno','col_cov','row_cov','G'])
    def Sr_DLrYLc_m_DWr_veiHiveWrDLrYLcWc_Wc(self):
        return self.covar.Sr()[:, sp.newaxis] * self.DLrYLc_m_DWr_veiHiveWrDLrYLcWc_Wc()

    @cached(['pheno','col_cov','row_cov','G'])
    def Rtilde_DLrYLcmDWrveiHiveWrDLrYLcWcWc_Ctilde(self, i):
        np_r = self.covar.Cr.getNumberParams()
        np_g = self.covar.Cg.getNumberParams()
        if i < np_r:
            R = self.WrWr_DLrYLc_m_DWr_veiHiveWrDLrYLcWc_Wc() 
        elif i < (np_r + np_g):
            R = self.Sr_DLrYLc_m_DWr_veiHiveWrDLrYLcWc_Wc() 
        else:
            R = self.DLrYLc_m_DWr_veiHiveWrDLrYLcWc_Wc()
        return sp.dot(R, self.covar.Ctilde(i))

    @cached(['row_cov', 'col_cov', 'pheno'])
    def Sr_DLrYLc_Ctilde(self, i):
        pass

    @cached(['row_cov', 'col_cov', 'designs', 'pheno'])
    def Sr_vei_dLWb_Ctilde(self, i):
        pass

    ######################
    # Areml
    ######################
    def Areml_K(self):
        pass

    def Areml_K_grad_i(self,i):
        pass

    ########################
    # LML terms
    ########################
    @cached(['row_cov', 'col_cov', 'designs', 'pheno'])
    def WKiy(self):
        pass

    @cached(['row_cov', 'col_cov', 'pheno'])
    def yKiy(self):
        r = (self.LrYLc()*self.DLrYLc()).sum()
        r-= (self.ve_WrDLrYLcWc() * self.Hi_ve_WrDLrYLcWc()).sum()
        return r

    @cached(['row_cov', 'col_cov', 'designs', 'pheno'])
    def yKiWb(self):
        pass

    #########################
    # Gradients
    #########################
    @cached(['row_cov', 'col_cov', 'pheno'])
    def yKiy_grad_i(self,i):
        return -sp.sum(self.DLrYLc_m_DWr_veiHiveWrDLrYLcWc_Wc() * self.Rtilde_DLrYLcmDWrveiHiveWrDLrYLcWcWc_Ctilde(i))

    @cached(['row_cov', 'col_cov', 'designs', 'pheno'])
    def yKiWb_grad_i(self,i):
        pass

    ############################
    # Simulate
    ############################
    def simulate_pheno(self):
        #region
        Z = sp.randn(self.covar.G.shape[1], self.covar.Cr.X.shape[1])
        Yr = sp.dot(self.covar.G, sp.dot(Z, self.covar.Cr.X.T))
        # background
        _S, _U = la.eigh(self.covar.Cg.K()); _S[_S<0] = 0
        Cg_h = _U*_S**0.5
        Rg_h = self.covar.Lr().T*(self.covar.Sr()**0.5)
        Yg = sp.dot(Rg_h, sp.dot(sp.randn(*self.mean.Y.shape), Cg_h.T))
        # noise
        _S, _U = la.eigh(self.covar.Cn.K()); _S[_S<0] = 0
        Cn_h = _U*_S**0.5
        Yn = sp.dot(sp.randn(*self.mean.Y.shape), Cn_h.T)
        RV = Yr+Yg+Yn
        return RV




import unittest
import numpy as np
from limix.core.covar import SQExpCov
from limix.core.covar import SumCov
from limix.utils.check_grad import mcheck_grad
import scipy as sp

class TestSumCov(unittest.TestCase):
    def setUp(self):
        np.random.seed(1)
        self._X1 = np.random.randn(10, 5)
        self._X2 = np.random.randn(10, 8)
        self._cov1 = SQExpCov(self._X1)
        self._cov2 = SQExpCov(self._X2)
        self._cov = SumCov(self._cov1, self._cov2)

    def test_sum_combination(self):
        K1 = self._cov1.K() + self._cov2.K()
        K2 = self._cov.K()

        np.testing.assert_almost_equal(K1, K2)

    def test_Kgrad(self):

        cov = self._cov

        def func(x, i):
            cov.setParams(x)
            return cov.K()

        def grad(x, i):
            cov.setParams(x)
            return cov.K_grad_i(i)

        x0 = cov.getParams()
        err = mcheck_grad(func, grad, x0)

        np.testing.assert_almost_equal(err, 0.)

    def test_Khess(self):

        cov = self._cov

        for j in range(cov.getNumberParams()):

            def func(x, i):
                cov.setParams(x)
                return cov.K_grad_i(j)

            def grad(x, i):
                cov.setParams(x)
                return cov.K_hess_i_j(j, i)

            x0 = cov.getParams()
            err = mcheck_grad(func, grad, x0)
            np.testing.assert_almost_equal(err, 0.)

    def test_use_to_predict_exception(self):
        with self.assertRaises(NotImplementedError):
            self._cov.use_to_predict = 1.

if __name__ == '__main__':
    unittest.main()



from .covar_base import Covariance
import pdb
import numpy as np
import scipy as sp
from .acombinators import ACombinatorCov
from hcache import Cached, cached

class SumCov(ACombinatorCov):
    """
    Sum of multiple covariance matrices.
    The number of paramteters is the sum of the parameters of the single covariances.
    """

    def __init__(self,*covars):
        """
        Args:
            covars:     covariances to be considered in the sum
        """
        ACombinatorCov.__init__(self)
        self.dim = None
        for covar in covars:
            self.addCovariance(covar)

    #####################
    # Cached
    #####################
    @cached('covar_base')
    def K(self):
        K = sp.zeros((self.dim,self.dim))
        for i in range(len(self.covars)):
            K += self.getCovariance(i).K()
        return K

    @cached('covar_base')
    def Kcross(self):
        R = None
        for i in range(len(self.covars)):
            if not self.getCovariance(i).use_to_predict:    continue
            if R is None:
                R = self.covars[i].Kcross()
            else:
                _ = self.covars[i].Kcross()
                assert _.shape[0]==R.shape[0], 'Dimension mismatch.'
                assert _.shape[1]==R.shape[1], 'Dimension mismatch.'
                R += _
        return R

    @cached('covar_base')
    def K_grad_i(self,i):
        istart = 0
        for j in range(len(self.covars)):
            istop = istart + self.getCovariance(j).getNumberParams()
            if (i < istop):
                idx = i - istart
                return self.getCovariance(j).K_grad_i(idx)
            istart = istop
        return None

    @cached('covar_base')
    def K_hess_i_j(self, i, j):
        istart = 0
        jstart = 0
        for c1 in range(len(self.covars)):
            istop = istart + self.getCovariance(c1).getNumberParams()
            if (i < istop):
                i0 = i - istart
                break
            istart = istop
        for c2 in range(len(self.covars)):
            jstop = jstart + self.getCovariance(c2).getNumberParams()
            if (j < jstop):
                j0 = j - jstart
                break
            jstart = jstop
        if c1==c2:
            r = self.getCovariance(c1).K_hess_i_j(i0, j0)
        else:
            r = sp.zeros((self.dim, self.dim))
        return r

    def _calcNumberParams(self):
        self.n_params = 0
        for i in range(len(self.covars)):
            self.n_params += self.getCovariance(i).getNumberParams()
        return self.n_params

    ####################
    # Interpretable Params
    ####################
    def K_grad_interParam_i(self,i):
        istart = 0
        for j in range(len(self.covars)):
            istop = istart + self.getCovariance(j).getNumberParams()
            if (i < istop):
                idx = i - istart
                return self.getCovariance(j).K_grad_interParam_i(idx)
            istart = istop
        return None





import scipy as sp
import scipy.linalg as la
import pdb

def decompose_GxE(Cr):
    _S, _U = la.eigh(Cr)
    _w = sp.sqrt(_S[-1]) * _U[:,[-1]]
    RV = {}
    RV['r_full'] = Cr 
    RV['r_block'] = sp.mean(Cr) * sp.ones(_U.shape)
    RV['r_rank1'] = sp.dot(_w, _w.T)
    return RV

def msqrt(M):
    S, U = la.eigh(M)
    S[S<0] = 0
    Mh = U * S**(0.5)
    return Mh

def var_CoXX(C, X):
    C_h = msqrt(C)
    PK = sp.kron(C_h, X)
    PK-= PK.mean(0)
    rv = (PK**2).sum() / float(PK.shape[0] - 1.)
    return rv

def calc_emp_pv_eff(xt, x0):
    idxt = sp.argsort(xt)[::-1]
    idx0 = sp.argsort(x0)[::-1]
    xts = xt[idxt]
    x0s = x0[idx0]
    it = 0
    i0 = 0
    _count = 0
    count = sp.zeros(xt.shape[0])
    while 1:
        if x0s[i0]>xts[it]:
            _count += 1
            i0+=1
            if i0==x0.shape[0]:
                count[idxt[it:]] = _count
                break
        else:
            count[idxt[it]] = _count
            it+=1
            if it==xt.shape[0]:
                break
    RV = (count + 1) / float(x0.shape[0])
    RV[RV > 1.] = 1.
    return RV




__author__ = 'damienarnol1'

import pdb

import numpy as np

from limix.core.covar import SQExpCov
from limix.core.covar import ZKZCov
from limix.core.covar import FixedCov
from limix.core.covar import SumCov
from limix.core.gp import GP
from limix.core.mean import MeanBase
from limix.utils.preprocess import covar_rescaling_factor


def normalise(mat, axis=1):
    mat -= np.reshape(mat.mean(axis=axis), [mat.shape[0], 1])
    mat /= np.reshape(mat.std(axis=axis), [mat.shape[0], 1])
    mat /= np.sqrt(mat.shape[1])
    return mat

rm_diag = True

data_directory = '/Users/damienarnol1/Documents/local/pro/PhD/data/Zurich_data/output/tmp/'

X_file = data_directory + '/cell_locations_1'
phenotype_file = data_directory+'/membrane_1'

# read data
X = np.genfromtxt(X_file, delimiter=',', skiprows=1)
phenotypes = np.genfromtxt(phenotype_file, delimiter=' ', skiprows=1)
phenotypes = np.delete(phenotypes, 0, axis=1)
phenotypes = normalise(phenotypes)
N_cells = phenotypes.shape[0]

Kinship = phenotypes.dot(phenotypes.transpose())
Kinship -= np.linalg.eigvalsh(Kinship).min() * np.eye(N_cells)
Kinship *= covar_rescaling_factor(Kinship)

phenotype = phenotypes[:, 5]
phenotype -= phenotype.mean()
phenotype /= phenotype.std()
phenotype = np.reshape(phenotype, [N_cells, 1])

# create different models and print the result including likelihood
# create all the covariance terms
direct_cov = FixedCov(Kinship)

# noise
noise_cov = FixedCov(np.eye(N_cells))

# local_noise
local_noise_cov = SQExpCov(X)

# environment effect
environment_cov = ZKZCov(X, Kinship, rm_diag)

# mean term
mean = MeanBase(phenotype)

#######################################################################
# MODEL 1: all
#######################################################################
print('...........................................................')
print('model 1 : complete model ')
print('...........................................................')
# total cov and mean
cov = SumCov(direct_cov, noise_cov)
cov = SumCov(cov, local_noise_cov)
cov = SumCov(cov, environment_cov)

# fixing length scale of ZKZ and SE
environment_cov.length = N_cells / 50
environment_cov.act_length = False

# local_noise_cov.length = N_cells/10.0
# local_noise_cov.act_length = False

# define and optimise GP
gp = GP(covar=cov, mean=mean)
gp.optimize()

# show results
print("inferred parameters ")
print("direct_scale = ", " ", direct_cov.scale)
print("noise_scale = ", " ", noise_cov.scale)
print("local_noise_scale = ", " ", local_noise_cov.scale)
print("local_noise_length = ", " ", local_noise_cov.length)
print("environment_scale = ", " ", environment_cov.scale)
print("environment_length = ", " ", environment_cov.length)

#######################################################################
# MODEL 2: no social effect
#######################################################################
print('...........................................................')
print('model 2: no social effect')
print('...........................................................')

direct_cov2 = FixedCov(Kinship)

# noise
noise_cov2 = FixedCov(np.eye(N_cells))

# local_noise
local_noise_cov2 = SQExpCov(X)

# total cov and mean
cov2 = SumCov(direct_cov2, noise_cov2)
cov2 = SumCov(cov2, local_noise_cov2)


# define and optimise GP
gp2 = GP(covar=cov2, mean=mean)
gp2.optimize()

# show results
print("inferred parameters ")
print("direct_scale = ", " ", direct_cov2.scale)
print("noise_scale = ", " ", noise_cov2.scale)
print("local_noise_scale = ", " ", local_noise_cov2.scale)
print("local_noise_length = ", " ", local_noise_cov2.length)

#######################################################################
# MODEL 3: no local moise no social effect
#######################################################################
print('...........................................................')
print('model 3: no local noise, no social effect')
print('...........................................................')
direct_cov3 = FixedCov(Kinship)

# noise
noise_cov3 = FixedCov(np.eye(N_cells))

# total cov and mean
cov3 = SumCov(direct_cov3, noise_cov3)


# define and optimise GP
gp3 = GP(covar=cov3, mean=mean)
gp3.optimize()

# show results
print("inferred parameters ")
print("direct_scale = ", " ", direct_cov3.scale)
print("noise_scale = ", " ", noise_cov3.scale)

#######################################################################
# MODEL 4: no local moise but social effect ON
#######################################################################
print('...........................................................')
print('model 3: no local noise, social effects ON')
print('...........................................................')
direct_cov4 = FixedCov(Kinship)

# noise
noise_cov4 = FixedCov(np.eye(N_cells))

# environment effect
environment_cov4 = ZKZCov(X, Kinship, rm_diag)

cov4 = SumCov(direct_cov4, noise_cov4)
cov4 = SumCov(cov4, environment_cov4)

# define and optimise GP
gp4 = GP(covar=cov4, mean=mean)
gp4.optimize()

# show results
print("inferred parameters ")
print("direct_scale = ", " ", direct_cov4.scale)
print("noise_scale = ", " ", noise_cov4.scale)
print("environment_scale = ", " ", environment_cov4.scale)
print("environment_length = ", " ", environment_cov4.length)


import sys
from limix.core.mean import MeanKronSum
from limix.core.covar import Cov2KronSumLR
from limix.core.covar import Covariance
from hcache import Cached, cached

import pdb
import numpy as NP
import scipy as sp
import scipy.linalg as LA
import time as TIME
from limix.core.gp import GP
from limix.core.covar.cov_reml import cov_reml
from limix.utils.util_functions import vec
from limix.core.utils import assert_type_or_list_type
from limix.core.utils import assert_type
from limix.core.utils import assert_subtype


class GP2KronSumLR(GP):
    """
    Gaussian Process with a 2kronSumLR Covariance and a mean that is a sum of Kronecker products:
        vec(Y) ~ N( vec( \sum_i F_i B_i A_i), Cr \kron GG.T + Cn \kron I )
    The current implementation supports only 1 fixed effect term!
    Notation:
        N = number of samples
        P = number of traits
        Y = [N, P] phenotype matrix
        F_i = sample fixed effect design for term i
        A_i = trait fixed effect design for term i
        B_i = effect sizes of fixed effect term i
        Cr = column covariance matrix for low-rank term
        Cn = column covariance matrix for noise term
        rank_c = rank of low-rank col covariance
        rank_r = rank of low-rank row covariance
    """

    def __init__(self, Y, Cn, G, F=None, A=None, rank=1, Cr=None):
        """
        Args:
            Y:      [N, P] phenotype matrix
            F:      Sample fixed effect design (first dimension must be N)
            A:      Trait fixed effect design (second dimension must be P)
            Cn:     Limix covariance matrix for Cn (dimension P)
            G:      [N, rank_r] numpy covariance matrix for G
            rank:   rank of column low-rank covariance (default = 1)
        """
        assert_type(Y, NP.ndarray, 'Y')
        assert_subtype(Cn, Covariance, 'Cn')
        assert_type(G, NP.ndarray, 'G')

        covar = Cov2KronSumLR(Cn=Cn, G=G, rank=rank, Cr=Cr)
        mean = MeanKronSum(Y=Y, F=F, A=A)
        assert mean.n_terms <= 1, ('GP2KronSum supports MeanKronSum'
                                   ' means with maximum 1 term!')
        GP.__init__(self, covar=covar, mean=mean)

    def _observe(self):
        self.covar.register(self.row_cov_has_changed, 'row_cov')
        self.covar.register(self.col_cov_has_changed, 'col_cov')
        self.mean.register(self.pheno_has_changed, 'pheno')
        self.mean.register(self.designs_have_changed, 'designs')

    def col_cov_has_changed(self):
        self.clear_cache('col_cov')
        self.clear_all()

    def row_cov_has_changed(self):
        self.clear_cache('row_cov')
        self.clear_all()

    def pheno_has_changed(self):
        self.clear_cache('pheno')
        self.clear_all()

    def designs_have_changed(self):
        self.clear_cache('designs')
        self.clear_all()

    ######################
    # Transformed phenotype
    ######################
    @cached(['pheno'])
    def YY(self):
        return sp.dot(self.mean.Y.T, self.mean.Y)

    @cached(['designs', 'pheno'])
    def FY(self):
        return sp.dot(self.mean.F[0].T, self.mean.Y)

    @cached(['row_cov', 'pheno'])
    def WrY(self):
        return sp.dot(self.covar.Wr(), self.mean.Y)

    @cached(['row_cov', 'pheno'])
    def GGY(self):
        return sp.dot(self.covar.G, sp.dot(self.covar.G.T, self.mean.Y))

    @cached(['row_cov', 'pheno'])
    def YGGY(self):
        return sp.dot(self.mean.Y.T, self.GGY()) 

    @cached(['row_cov', 'pheno'])
    def WrGGY(self):
        return sp.dot(self.covar.Wr(), self.GGY())

    @cached(['col_cov', 'pheno'])
    def YLc(self):
        return sp.dot(self.mean.Y, self.covar.Lc().T)

    @cached(['row_cov', 'col_cov', 'pheno'])
    def WrYLc(self):
        return sp.dot(self.WrY(), self.covar.Lc().T)

    @cached(['row_cov', 'col_cov', 'pheno'])
    def WrYLcWc(self):
        return sp.dot(self.WrYLc(), self.covar.Wc().T)

    @cached(['designs', 'col_cov', 'pheno'])
    def FYLc(self):
        return sp.dot(self.FY(), self.covar.Lc().T)

    @cached(['designs', 'col_cov', 'pheno'])
    def FYLcALc(self):
        return sp.dot(self.FYLc(), self.ALc().T)

    @cached(['row_cov', 'col_cov', 'pheno'])
    def GGYLc(self):
        return sp.dot(self.GGY(), self.covar.Lc().T)

    @cached(['row_cov', 'col_cov', 'pheno'])
    def WrGGYLc(self):
        return sp.dot(self.WrGGY(), self.covar.Lc().T)

    @cached(['row_cov', 'col_cov', 'pheno'])
    def DWrYLcWc(self):
        return self.covar.D() * self.WrYLcWc()

    #@cached(['row_cov', 'col_cov', 'pheno'])
    #def RYLcCtilde(self, i):
    #    if i < self.covar.Cr.getNumberParams():
    #        RYLc = self.GGYLc()
    #    else:
    #        RYLc = self.YLc()
    #    return sp.dot(RYLc, self.covar.Ctilde(i))

    @cached(['row_cov', 'col_cov', 'pheno'])
    def SrDWrYLcWcCbar(self, i):
        if i < self.covar.Cr.getNumberParams():
            SgDWrYLcWc = self.covar.Sg()[:, sp.newaxis] *  self.DWrYLcWc()
        else:
            SgDWrYLcWc = self.DWrYLcWc()
        return sp.dot(SgDWrYLcWc, self.covar.Cbar(i))

    def YRY(self, i):
        if i < self.covar.Cr.getNumberParams():
            return self.YGGY()
        else:
            return self.YY()

    @cached(['row_cov', 'col_cov', 'pheno'])
    def WrRYLcCtildeWc(self, i):
        if i < self.covar.Cr.getNumberParams():
            WrRYLc = self.WrGGYLc()
        else:
            WrRYLc = self.WrYLc()
        return sp.dot(WrRYLc, sp.dot(self.covar.Ctilde(i), self.covar.Wc().T))

    ###############################################
    # Fixed effects
    # uses that MeanKronSum has only one term
    ###############################################
    @cached(['designs', 'row_cov'])
    def WrF(self):
        return sp.dot(self.covar.Wr(), self.mean.F[0])

    @cached(['designs', 'row_cov'])
    def GGF(self):
        return sp.dot(self.covar.G, sp.dot(self.covar.G.T, self.mean.F[0]))

    @cached(['designs', 'row_cov'])
    def YGGF(self):
        return sp.dot(self.mean.Y.T, self.GGF()) 

    @cached(['designs', 'row_cov'])
    def FGGF(self):
        return sp.dot(self.mean.F[0].T, self.GGF())

    @cached('designs')
    def FF(self):
        return sp.dot(self.mean.F[0].T, self.mean.F[0])

    @cached(['designs', 'row_cov'])
    def WrGGF(self):
        return sp.dot(self.covar.Wr(), self.GGF())

    @cached(['designs', 'row_cov', 'col_cov', 'pheno'])
    def FB(self):
        return sp.dot(self.mean.F[0], self.mean.B[0])

    @cached(['designs', 'row_cov', 'col_cov', 'pheno'])
    def YGGFBA(self):
        return sp.dot(sp.dot(self.YGGF(), self.mean.B[0]), self.mean.A[0]) 

    @cached(['designs', 'row_cov', 'col_cov', 'pheno'])
    def YFBA(self):
        return sp.dot(sp.dot(self.FY().T, self.mean.B[0]), self.mean.A[0]) 

    @cached(['designs', 'row_cov', 'col_cov', 'pheno'])
    def WrFB(self):
        return sp.dot(self.WrF(), self.mean.B[0])

    @cached(['designs', 'row_cov', 'col_cov', 'pheno'])
    def GGFB(self):
        return sp.dot(self.GGF(), self.mean.B[0])

    @cached(['designs', 'row_cov', 'col_cov', 'pheno'])
    def WrGGFB(self):
        return sp.dot(self.WrGGF(), self.mean.B[0])

    @cached(['designs', 'row_cov', 'col_cov', 'pheno'])
    def FBALc(self):
        return sp.dot(self.FB(), self.ALc())

    @cached(['designs', 'row_cov', 'col_cov', 'pheno'])
    def WrFBALc(self):
        return sp.dot(self.WrFB(), self.ALc())

    @cached(['designs', 'row_cov', 'col_cov', 'pheno'])
    def GGFBALc(self):
        return sp.dot(self.GGFB(), self.ALc())

    @cached(['designs', 'row_cov', 'col_cov', 'pheno'])
    def WrGGFBALc(self):
        return sp.dot(self.WrGGFB(), self.ALc())

    @cached(['designs', 'row_cov', 'col_cov', 'pheno'])
    def WrGGFBALcWc(self):
        return sp.dot(self.WrGGFBALc(), self.covar.Wc().T)

    @cached(['designs', 'row_cov', 'col_cov', 'pheno'])
    def WrFBALcWc(self):
        return sp.dot(self.WrFBALc(), self.covar.Wc().T)

    @cached(['designs', 'row_cov', 'col_cov', 'pheno'])
    def DWrGGFBALcWc(self):
        return self.covar.D() * self.WrGGFBALcWc()

    @cached(['designs', 'row_cov', 'col_cov', 'pheno'])
    def DWrFBALcWc(self):
        return self.covar.D() * self.WrFBALcWc()

    @cached(['designs', 'col_cov'])
    def ALc(self):
        return sp.dot(self.mean.A[0], self.covar.Lc().T)

    @cached(['designs', 'col_cov'])
    def ALcLcA(self):
        return sp.dot(self.ALc(), self.ALc().T)

    @cached(['designs', 'col_cov'])
    def ALcWc(self):
        return sp.dot(self.ALc(), self.covar.Wc().T)

    @cached(['designs', 'row_cov', 'col_cov'])
    def WLW(self):
        # confusing notation the first W refers to kroneckered matrix
        # that brings to the low-dimensional space while the second refers to the fixed effect design
        return sp.kron(self.ALcWc().T, self.WrF())

    @cached(['designs', 'row_cov', 'col_cov'])
    def dWLW(self):
        return self.covar.d()[:,sp.newaxis] * self.WLW()

    #@cached(['designs', 'row_cov', 'col_cov', 'pheno'])
    #def RFBALcCtilde(self, i):
    #    if i < self.covar.Cr.getNumberParams():
    #        RFBALc = self.GGFBALc()
    #    else:
    #        RFBALc = self.FBALc()
    #    return sp.dot(RFBALc, self.covar.Ctilde(i))

    def YRFBA(self, i):
        if i < self.covar.Cr.getNumberParams():
            return self.YGGFBA()
        else:
            return self.YFBA()

    @cached(['designs', 'row_cov', 'col_cov', 'pheno'])
    def SrDWrFBALcWcCbar(self, i):
        if i < self.covar.Cr.getNumberParams():
            SgDWrFBALcWc = self.covar.Sg()[:, sp.newaxis] *  self.DWrFBALcWc()
        else:
            SgDWrFBALcWc = self.DWrFBALcWc()
        return sp.dot(SgDWrFBALcWc, self.covar.Cbar(i))

    @cached(['designs', 'row_cov', 'col_cov', 'pheno'])
    def WrRFBALcCtildeWc(self, i):
        if i < self.covar.Cr.getNumberParams():
            WrRFBALc = self.WrGGFBALc()
        else:
            WrRFBALc = self.WrFBALc()
        return sp.dot(WrRFBALc, sp.dot(self.covar.Ctilde(i), self.covar.Wc().T))

    # fixed effect methods for Areml_grad

    @cached('col_cov')
    def CtildeLcA(self, i):
        return sp.dot(self.covar.Ctilde(i), self.ALc().T)

    @cached(['designs', 'row_cov', 'col_cov'])
    def ALcCtildeLcA_o_FRF(self, i):
        if i < self.covar.Cr.getNumberParams():
            FRF = self.FGGF()
        else:
            FRF = self.FF()
        ALcCtildeALc = sp.dot(self.ALc(), self.CtildeLcA(i))
        return sp.kron(ALcCtildeALc, FRF)

    @cached(['designs', 'row_cov', 'col_cov'])
    def Cbar_o_Sr_dWLW(self, i):
        if i < self.covar.Cr.getNumberParams():
            Cbar_o_Sr = sp.kron(self.covar.Cbar(i), sp.diag(self.covar.Sg()))
        else:
            Cbar_o_Sr = sp.kron(self.covar.Cbar(i), sp.eye(self.covar.rank_r))
        return sp.dot(Cbar_o_Sr, self.dWLW())

    @cached(['designs', 'row_cov', 'col_cov'])
    def WcCtildeLcA_o_WrRF(self, i):
        if i < self.covar.Cr.getNumberParams():
            self.WrRF = self.WrGGF()
        else:
            self.WrRF = self.WrF()
        WcCtildeLcA = sp.dot(self.covar.Wc(), self.CtildeLcA(i))
        return sp.kron(WcCtildeLcA, self.WrRF)


    ######################
    # Areml
    ######################
    def Areml_K(self):
        R  = sp.kron(self.ALcLcA(), self.FF())
        R -= sp.dot(self.WLW().T, self.dWLW())
        return R

    def Areml_K_grad_i(self,i):
        i = self.covar._actindex2index(i)
        R = sp.dot(self.WcCtildeLcA_o_WrRF(i).T, self.dWLW())
        R+= R.T
        R+= -self.ALcCtildeLcA_o_FRF(i)
        R+= -sp.dot(self.dWLW().T, self.Cbar_o_Sr_dWLW(i))
        return R

    ########################
    # LML terms
    ########################
    @cached(['designs', 'row_cov', 'col_cov', 'pheno'])
    def WKiy(self):
        R = vec(self.FYLcALc())
        # the following could be further optimized but do not care for now
        R -= sp.dot(self.WLW().T, vec(self.DWrYLcWc()))
        return R

    @cached(['row_cov', 'col_cov', 'designs', 'pheno'])
    def b(self):
        if self.mean.n_covs > 0:
            R = self.Areml.solve(self.WKiy())
        else:
            R = None
        return R

    @cached(['row_cov', 'col_cov', 'pheno'])
    def yKiy(self):
        return (self.YLc() * self.YLc()).sum() - (self.WrYLcWc() * self.DWrYLcWc()).sum()

    @cached(['designs', 'row_cov', 'col_cov', 'pheno'])
    def yKiWb(self):
        return (self.WKiy() * self.mean.b).sum()

    #########################
    # Gradients
    #########################
    @cached(['row_cov', 'col_cov', 'pheno'])
    def yKiy_grad_i(self,i):
        i = self.covar._actindex2index(i)
        r = - (self.YRY(i).T * self.covar.LcCtildeLc(i)).sum()
        r+= - (self.DWrYLcWc() * self.SrDWrYLcWcCbar(i)).sum()
        r+= 2 * (self.WrRYLcCtildeWc(i) * self.DWrYLcWc()).sum()
        return r

    @cached(['designs', 'row_cov', 'col_cov', 'pheno'])
    def yKiWb_grad_i(self,i):
        Areml_grad_i_b = sp.dot(self.Areml.K_grad_i(i), self.mean.b) 
        i = self.covar._actindex2index(i)
        #r = -2 * (self.YLc() * self.RFBALcCtilde(i)).sum()
        r = - 2 * (self.YRFBA(i).T * self.covar.LcCtildeLc(i)).sum()
        r+= -2 * (self.DWrYLcWc() * self.SrDWrFBALcWcCbar(i)).sum()
        r+= 2 * (self.WrRYLcCtildeWc(i) * self.DWrFBALcWc()).sum()
        r+= 2 * (self.WrRFBALcCtildeWc(i) * self.DWrYLcWc()).sum()
        r-= (self.mean.b * Areml_grad_i_b).sum()
        #r+= (self.FBALc() * self.RFBALcCtilde(i)).sum()
        #r+= (self.DWrFBALcWc() * self.SrDWrFBALcWcCbar(i)).sum()
        #r+= - 2 * (self.WrRFBALcCtildeWc(i) * self.DWrFBALcWc()).sum()
        return r

    ############################
    # Simulate
    ############################
    def simulate_pheno(self):
        Yc = sp.dot(self.mean.F[0], sp.dot(self.mean.B[0], self.mean.A[0].T))
        Z = sp.randn(self.covar.G.shape[1], self.covar.Cr.X.shape[1])
        Yr = sp.dot(self.covar.G, sp.dot(Z, self.covar.Cr.X.T))
        _S, _U = LA.eigh(self.covar.Cn.K()); _S[_S<0] = 0
        Cn_h = _U*_S**0.5
        Yn = sp.dot(sp.randn(*self.mean.Y.shape), Cn_h.T)
        RV = Yc+Yr+Yn
        return RV
    



import sys
from limix.core.mean import MeanKronSum
from limix.core.covar import Cov2KronSum
from limix.core.covar import Covariance
from hcache import Cached, cached

import pdb
import numpy as NP
import scipy as sp
import scipy.linalg as LA
import time as TIME
from .gp_base import GP
from limix.core.covar.cov_reml import cov_reml
from limix.core.utils import assert_type_or_list_type
from limix.core.utils import assert_type
from limix.core.utils import assert_subtype


class GP2KronSum(GP):
    """
    Gaussian Process with a 2kronSum Covariance and a mean that is a sum of Kronecker products:
        vec(Y) ~ N( vec( \sum_i F_i B_i A_i), Cg \kron R + Cn \kron I )
    Notation:
        N = number of samples
        P = number of traits
        Y = [N, P] phenotype matrix
        F_i = sample fixed effect design for term i
        A_i = trait fixed effect design for term i
        B_i = effect sizes of fixed effect term i
        Cg = column covariance matrix for signal term respectively
        Cn = column covariance matrix for noise term respectively
        R = row covariance matrix for signal term respectively
    """

    def __init__(self, Y, Cg, Cn, F=None, A=None, R=None, S_R=None, U_R=None):
        """
        Args:
            Y:      [N, P] phenotype matrix
            F:      list of sample fixed effect designs.
                    Each term must have first dimension N
            A:      list of trait fixed effect design.
                    Each term must have second dimension P
            Cg:     Limix covariance matrix for Cg (dimension P)
            Cn:     Limix covariance matrix for Cn (dimension P)
            R:      [N, N] numpy semidemidefinite covariance matrix for R.
                    In alternative to R, S_R and U_R can be specified.
            S_R:    N vector of eigenvalues of R
            U_R:    [N, N] eigenvector matrix of R
        """
        assert_type(Y, NP.ndarray, 'Y')
        assert_subtype(Cg, Covariance, 'Cg')
        assert_subtype(Cn, Covariance, 'Cn')

        covar = Cov2KronSum(Cg=Cg, Cn=Cn, R=R, S_R=S_R, U_R=U_R)
        mean = MeanKronSum(Y=Y, F=F, A=A)

        GP.__init__(self, covar=covar, mean=mean)

    def _observe(self):
        self.covar.register(self.row_cov_has_changed, 'row_cov')
        self.covar.register(self.col_cov_has_changed, 'col_cov')
        self.mean.register(self.pheno_has_changed, 'pheno')
        self.mean.register(self.designs_have_changed, 'designs')

    def col_cov_has_changed(self):
        self.clear_cache('col_cov')
        self.clear_all()

    def row_cov_has_changed(self):
        self.clear_cache('row_cov')
        self.clear_all()

    def pheno_has_changed(self):
        self.clear_cache('pheno')
        self.clear_all()

    def designs_have_changed(self):
        self.clear_cache('designs')
        self.clear_all()


    ######################
    # Transformed phenotype
    ######################
    @cached('row_cov', 'pheno')
    def LrY(self):
        return sp.dot(self.covar.Lr(), self.mean.Y)

    @cached(['row_cov', 'col_cov', 'pheno'])
    def LrYLc(self):
        return sp.dot(self.LrY(), self.covar.Lc().T)

    @cached(['row_cov', 'col_cov', 'pheno'])
    def DLrYLc(self):
        return self.covar.D() * self.LrYLc()

    ######################
    # Transformed fixed effects
    ######################
    @cached(['row_cov', 'designs'])
    def LrF(self):
        R = []
        for ti in range(self.mean.n_terms):
            R.append(sp.dot(self.covar.Lr(), self.mean.F[ti]))
        return R

    @cached(['col_cov', 'designs'])
    def ALc(self):
        R = []
        for ti in range(self.mean.n_terms):
            R.append(sp.dot(self.mean.A[ti], self.covar.Lc().T))
        return R

    @cached(['row_cov', 'col_cov', 'designs'])
    def LW(self):
        R = sp.zeros((self.mean.Y.size, self.mean.n_covs))
        istart = 0
        for ti in range(self.mean.n_terms):
            iend = istart + self.mean.F[ti].shape[1] * self.mean.A[ti].shape[0]
            R[:, istart:iend] = sp.kron(self.ALc()[ti].T, self.LrF()[ti])
            istart += iend
        return R

    @cached(['row_cov', 'col_cov', 'designs'])
    def dLW(self):
        return self.covar.d()[:,sp.newaxis] * self.LW()

    @cached(['row_cov', 'col_cov', 'pheno'])
    def Sr_DLrYLc_Ctilde(self, i):
        return self.covar.Sr_X_Ctilde(self.DLrYLc(), i)

    @cached(['row_cov', 'col_cov', 'designs', 'pheno'])
    def Sr_vei_dLWb_Ctilde(self, i):
        return self.covar.Sr_X_Ctilde(self.vei_dLWb(), i)

    @cached(['row_cov', 'col_cov', 'designs', 'pheno'])
    def vei_dLWb(self):
        # could be optimized but probably not worth it
        # as it requires a for loop
        r = sp.dot(self.dLW(), self.mean.b)
        return r.reshape(self.mean.Y.shape, order = 'F')

    ######################
    # Areml
    ######################
    def Areml_K(self):
        return sp.dot(self.LW().T, self.dLW())

    def Areml_K_grad_i(self,i):
        dLWt = self.dLW().reshape((self.mean._N, self.mean._P, self.mean.n_covs), order = 'F')
        if i < self.covar.Cg.getNumberParams():
            SrdLWt = self.covar.Sr()[:, sp.newaxis, sp.newaxis] * dLWt
        else:
            SrdLWt = dLWt
        SrdLWtC = sp.tensordot(SrdLWt, self.covar.Ctilde(i), axes=(1, 1))
        SroCdLW = SrdLWtC.swapaxes(1,2).reshape((self.mean._N * self.mean._P, self.mean.n_covs), order = 'F')
        return -sp.dot(self.dLW().T, SroCdLW)

    ########################
    # LML terms
    ########################
    @cached(['row_cov', 'col_cov', 'designs', 'pheno'])
    def WKiy(self):
        R = sp.zeros((self.mean.n_covs, 1))
        istart = 0
        for ti in range(self.mean.n_terms):
            _dim = self.mean.F[ti].shape[1] * self.mean.A[ti].shape[0]
            iend = istart + _dim
            FLrDLrYLc = sp.dot(self.LrF()[ti].T, self.DLrYLc())
            R[istart:iend, 0] = sp.dot(FLrDLrYLc, self.ALc()[ti].T).reshape(_dim, order = 'F')
        return R

    @cached(['row_cov', 'col_cov', 'designs', 'pheno'])
    def b(self):
        if self.mean.n_covs > 0:
            R = self.Areml.solve(self.WKiy())
        else:
            R = None
        return R

    @cached(['row_cov', 'col_cov', 'pheno'])
    def yKiy(self):
        return (self.LrYLc()*self.DLrYLc()).sum()

    @cached(['row_cov', 'col_cov', 'designs', 'pheno'])
    def yKiWb(self):
        return (self.LrYLc() * self.vei_dLWb()).sum()

    #########################
    # Gradients
    #########################
    @cached(['row_cov', 'col_cov', 'pheno'])
    def yKiy_grad_i(self,i):
        return -(self.DLrYLc() * self.Sr_DLrYLc_Ctilde(i)).sum()

    @cached(['row_cov', 'col_cov', 'designs', 'pheno'])
    def yKiWb_grad_i(self,i):
        rv = -2*(self.DLrYLc()*self.Sr_vei_dLWb_Ctilde(i)).sum()
        rv+= (self.vei_dLWb()*self.Sr_vei_dLWb_Ctilde(i)).sum()
        return rv

    ############################
    # Simulate
    ############################
    def simulate_pheno(self, Rh=None):
        if Rh is None:
            S, U = la.eigh(self.covar.R)
            S[S<0] = 0
            Rh = U*S**(0.5)
        Yc = sp.dot(self.mean.F[0], sp.dot(self.mean.B[0], self.mean.A[0].T))
        Z = sp.randn(Rh.shape[1], self.covar.Cr.X.shape[1])
        Yr = sp.dot(Rh, sp.dot(Z, self.covar.Cr.X.T))
        _S, _U = LA.eigh(self.covar.Cn.K()); _S[_S<0] = 0
        Cn_h = _U*_S**0.5
        Yn = sp.dot(sp.randn(*self.mean.Y.shape), Cn_h.T)
        RV = Yc+Yr+Yn
        return RV




import pdb
import scipy as sp
import scipy.linalg
import copy
import sys
import time
from limix.core.type.observed import Observed
from hcache import Cached, cached
from limix.core.covar import Covariance
from limix.core.mean import MeanBase
from limix.core.covar.cov_reml import cov_reml
import limix.core.optimize.optimize_bfgs as OPT
from .relay import GPMeanRelay

import logging
logger = logging.getLogger(__name__)


class GP(Cached, Observed):
    """
    Gaussian Process regression class for linear mean (with REML)
    y ~ N(Wb,K)
    Notation:
        N = number of samples
        K = number of fixed effects
        y = phenotype vector
        W = fixed effect design
        K = covariance function
    """

    def __init__(self, mean, covar):
        """
        covar:        Limix covariance function
        mean:         Limix linear Mean function
        """
        Cached.__init__(self)

        if not issubclass(type(mean), MeanBase):
            raise TypeError('Parameter mean must have MeanBase inheritance.')

        if not issubclass(type(covar), Covariance):
            raise TypeError('Parameter covar must have Covariance '
                            'inheritance.')

        self.covar = covar
        self.mean = mean
        self.mean._set_relay(GPMeanRelay(self))
        self.Areml = cov_reml(self)
        self._observe()

    def _observe(self):
        # different notification should be possible
        # e.g. for mean: pheno and designs
        # see GP2KronSum
        self.covar.register(self.clear_all)
        self.mean.register(self.clear_all)

    def clear_all(self):
        self._notify() # notify Areml
        self.clear_cache('gp_base')

    def setParams(self, params):
        self.covar.setParams(params['covar'])

    def getParams(self):
        RV = {}
        RV['covar'] = self.covar.getParams()
        return RV

    ######################
    # Areml
    ######################
    def Areml_K(self):
        return sp.dot(self.mean.W.T, self.KiW())

    def Areml_K_grad_i(self, i):
        return -sp.dot(self.KiW().T, self.DiKKiW(i))

    #######################
    # LML terms
    #######################
    @cached('gp_base')
    def KiW(self):
        return self.covar.solve(self.mean.W)

    @cached('gp_base')
    def yKiW(self):
        return sp.dot(self.mean.y.T, self.KiW())

    @cached('gp_base')
    def b(self):
        if self.mean.n_covs > 0:
            R = self.Areml.solve(self.yKiW().T)
        else:
            R = None
        return R

    @cached('gp_base')
    def KiWb(self):
        return sp.dot(self.KiW(), self.mean.b)

    @cached('gp_base')
    def Kiy(self):
        return self.covar.solve(self.mean.y)

    @cached('gp_base')
    def yKiy(self):
        return (self.mean.y*self.Kiy()).sum()

    @cached('gp_base')
    def yKiWb(self):
        return (self.mean.y*self.KiWb()).sum()

    #######################
    # gradients
    #######################
    @cached('gp_base')
    def DiKKiy(self, i):
        return self.covar.K_grad_i_dot(self.Kiy(), i)

    @cached('gp_base')
    def DiKKiW(self, i):
        return self.covar.K_grad_i_dot(self.KiW(), i)

    @cached('gp_base')
    def DiKKiWb(self, i):
        return sp.dot(self.DiKKiW(i), self.mean.b)

    @cached('gp_base')
    def yKiy_grad_i(self, i):
        return -(self.Kiy()*self.DiKKiy(i)).sum()

    @cached('gp_base')
    def yKiWb_grad_i(self, i):
        rv = -2*(self.Kiy()*self.DiKKiWb(i)).sum()
        rv += (self.KiWb()*self.DiKKiWb(i)).sum()
        return rv

    #######################
    # LML and gradients
    #######################

    @cached('gp_base')
    def LML(self):
        # const term to add?
        rv = 0.5*self.covar.logdet()
        rv += 0.5*self.yKiy()
        if self.mean.n_covs > 0:
            rv += 0.5*self.Areml.logdet()
            rv -= 0.5*self.yKiWb()
        return rv

    @cached('gp_base')
    def yKiy_grad(self):
        n_params = self.getParams()['covar'].shape[0]
        RV = {'covar': sp.zeros(n_params)}
        for i in range(n_params):
            RV['covar'][i] = self.yKiy_grad_i(i)
        return RV

    @cached('gp_base')
    def yKiWb_grad(self):
        n_params = self.getParams()['covar'].shape[0]
        RV = {'covar': sp.zeros(n_params)}
        if self.mean.n_covs > 0:
            for i in range(n_params):
                RV['covar'][i] = self.yKiWb_grad_i(i)
        return RV

    @cached('gp_base')
    def Areml_logdet_grad(self):
        n_params = self.getParams()['covar'].shape[0]
        RV = {'covar': sp.zeros(n_params)}
        if self.mean.n_covs > 0:
            for i in range(n_params):
                RV['covar'][i] = self.Areml.logdet_grad_i(i)
        return RV

    @cached('gp_base')
    def LML_grad(self):
        n_params = self.getParams()['covar'].shape[0]
        RV = {'covar': sp.zeros(n_params)}
        for i in range(n_params):
            RV['covar'][i] = 0.5*self.covar.logdet_grad_i(i)
            RV['covar'][i] += 0.5*self.yKiy_grad_i(i)
            if self.mean.n_covs > 0:
                RV['covar'][i] += 0.5*self.Areml.logdet_grad_i(i)
                RV['covar'][i] -= 0.5*self.yKiWb_grad_i(i)
        return RV

    def predict(self):
        R = None
        if self.covar.use_to_predict:
            Kcross = self.covar.Kcross()
            Kiyres = self.Kiy()-self.KiWb()
            R = sp.dot(Kcross, Kiyres)
        if self.mean.use_to_predict:
            _ = self.mean.predict()
            if R is None:
                R = _
            else:
                assert _.shape[0] == R.shape[0], 'Dimension mismatch'
                assert _.shape[1] == R.shape[1], 'Dimension mismatch'
                R += _
        return R

    #########################
    # OPTIMIZATION
    #########################
    def optimize(self, calc_ste=False, Ifilter=None, bounds=None, verbose=True,
                 opts={}, *args, **kw_args):
        # logger.info('Marginal likelihood optimization.')

        if verbose:
            print('Marginal likelihood optimization.')
        t0 = time.time()
        conv, info = OPT.opt_hyper(self, Ifilter=Ifilter, bounds=bounds,
                                   opts=opts, *args, **kw_args)
        t1 = time.time()

        # if logger.levelno == logger.DEBUG:
        if verbose:
            # logger.debug('Time elapsed: %.2fs', t1-t0)
            print(('Converged:', conv))
            print(('Time elapsed: %.2f s' % (t1-t0)))
            grad = self.LML_grad()
            grad_norm = 0
            for key in list(grad.keys()):
                grad_norm += (grad[key]**2).sum()
            grad_norm = sp.sqrt(grad_norm)
            print(('Log Marginal Likelihood: %.7f.' % self.LML()))
            print(('Gradient norm: %.7f.' % grad_norm))
            # logger.debug('Log Marginal Likelihood: %.7f.', self.LML())
            # logger.debug('Gradient norm: %.7f.', grad_norm)

        if calc_ste:
            self.calc_ste(verbose=verbose)
        return conv, info

    def calc_ste(self, verbose=True):
        if verbose:
            print('Standard errors calculation.')
            # logger.info('Standard error calculation.')
        t0 = time.time()
        I_covar = self.covar.getFisherInf()
        I_mean = self.Areml.K()
        self.covar.setFIinv(sp.linalg.pinv(I_covar))
        if self.mean.n_covs>0:
            self.mean.setFIinv(sp.linalg.pinv(I_mean))
        t1 = time.time()
        # logger.debug('Time elapsed: %.2fs', t1-t0)
        if verbose:
            print(('Time elapsed: %.2f s' % (t1-t0)))


    def test_grad(self):
        from limix.utils.check_grad import mcheck_grad

        def func(x, i):
            params = self.getParams()
            params['covar'] = x
            self.setParams(params)
            return self.LML()

        def grad(x, i):
            params = self.getParams()
            params['covar'] = x
            self.setParams(params)
            grad = self.LML_grad()
            return grad['covar'][i]

        x0 = self.getParams()['covar']
        err = mcheck_grad(func, grad, x0)
        print(err)
        # np.testing.assert_almost_equal(err, 0., decimal=5)



import unittest
import limix 
import scipy as sp
import scipy.linalg as LA
import pdb
import os

dir_name = os.path.dirname(os.path.realpath(__file__))
base_folder = os.path.join(dir_name, 'data')

class TestMTSet(unittest.TestCase):
    """test class for optimization""" 
    
    def setUp(self):
        sp.random.seed(0)
        self.Y = sp.loadtxt(os.path.join(base_folder, 'Y.txt')) 
        self.XX = sp.loadtxt(os.path.join(base_folder, 'XX.txt')) 
        self.Xr = sp.loadtxt(os.path.join(base_folder, 'Xr.txt')) 
        self.N,self.P = self.Y.shape
        self.write = False 

    def test_mtSetNull(self):
        fbasename = 'mtSetNull'
        setTest = limix.MTSet(Y=self.Y, R=self.XX)
        nullMTInfo = setTest.fitNull(cache=False)
        ext = {'Cg':nullMTInfo['Cg'],'Cn':nullMTInfo['Cn']}
        if self.write: self.saveStuff(fbasename,ext)
        RV = self.assess(fbasename,ext)
        self.assertTrue(RV)

    def test_mtSetNull_eigenCache(self):
        fbasename = 'mtSetNull'
        S,U = LA.eigh(self.XX)
        setTest = limix.MTSet(Y=self.Y, S_R=S, U_R=U)
        nullMTInfo = setTest.fitNull(cache=False)
        ext = {'Cg':nullMTInfo['Cg'],'Cn':nullMTInfo['Cn']}
        if self.write: self.saveStuff(fbasename,ext)
        RV = self.assess(fbasename,ext)
        self.assertTrue(RV)

    def test_mtSet_base(self):
        fbasename = 'mtSet_base'
        setTest = limix.MTSet(self.Y, R= self.XX)
        optInfo = setTest.optimize(self.Xr)
        ext = {'Cr':optInfo['Cr'],
               'Cg':optInfo['Cg'],
               'Cn':optInfo['Cn']}
        if self.write: self.saveStuff(fbasename,ext)
        RV = self.assess(fbasename,ext)
        self.assertTrue(RV)

    def test_mtSet_eigenCache(self):
        fbasename = 'mtSet_base'
        S,U = LA.eigh(self.XX)
        setTest = limix.MTSet(Y=self.Y, S_R=S, U_R=U)
        optInfo = setTest.optimize(self.Xr)
        ext = {'Cr':optInfo['Cr'],
               'Cg':optInfo['Cg'],
               'Cn':optInfo['Cn']}
        RV = self.assess(fbasename,ext)
        self.assertTrue(RV)

    def test_mtSetPCnull(self):
        fbasename = 'mtSetPCnull'
        setTest = limix.MTSet(Y=self.Y)
        nullMTInfo = setTest.fitNull(cache=False)
        ext = {'Cn':nullMTInfo['Cn']}
        if self.write: self.saveStuff(fbasename,ext)
        RV = self.assess(fbasename,ext)
        self.assertTrue(RV)

    def test_mtSetPCnull_fixed(self):
        fbasename = 'mtSetPCnull_fixed'
        setTest = limix.MTSet(Y=self.Y, F=self.Xr)
        nullMTInfo = setTest.fitNull(cache=False)
        ext = {'Cg':nullMTInfo['Cg'],'Cn':nullMTInfo['Cn']}
        if self.write: self.saveStuff(fbasename,ext)
        RV = self.assess(fbasename,ext)
        self.assertTrue(RV)

    def test_mtSetPC_base(self):
        fbasename = 'mtSetPC_base'
        setTest = limix.MTSet(Y=self.Y)
        optInfo = setTest.optimize(self.Xr)
        ext = {'Cr':optInfo['Cr'],
               'Cn':optInfo['Cn']}
        if self.write: self.saveStuff(fbasename,ext)
        RV = self.assess(fbasename,ext)
        self.assertTrue(RV)

    def test_mtSetPC_fixed(self):
        fbasename = 'mtSetPC_fixed'
        setTest = limix.MTSet(Y=self.Y, F=self.Xr[:,:2])
        optInfo = setTest.optimize(self.Xr)
        ext = {'Cr':optInfo['Cr'],
               'Cn':optInfo['Cn']}
        if self.write: self.saveStuff(fbasename,ext)
        RV = self.assess(fbasename,ext)
        self.assertTrue(RV)

    def saveStuff(self,fbasename,ext):
        """ util function """ 
        base = os.path.join(base_folder, 'res_'+fbasename+'_')
        for key in list(ext.keys()): 
            sp.savetxt(base+key+'.txt',ext[key])

    def loadStuff(self,fbasename,keys):
        """ util function """ 
        RV = {}
        base = os.path.join(base_folder, 'res_'+fbasename+'_')
        for key in keys: 
            RV[key] = sp.loadtxt(base+key+'.txt')
        return RV

    def assess(self,fbasename,ext):
        """ returns a bool vector """
        real = self.loadStuff(fbasename,list(ext.keys())) 
        RV = sp.all([((ext[key]-real[key])**2).mean()<1e-6 for key in list(ext.keys())])
        if not RV:  pdb.set_trace()
        return RV

if __name__ == '__main__':
    unittest.main()




import sys
from limix.core.old.cobj import *
from limix.utils.preprocess import regressOut
import numpy as np

import scipy.linalg as LA
import copy

def compute_X1KX2(Y, D, X1, X2, A1=None, A2=None):

    R,C = Y.shape
    if A1 is None:
        nW_A1 = Y.shape[1]
        #A1 = np.eye(Y.shape[1])	#for now this creates A1 and A2
    else:
        nW_A1 = A1.shape[0]

    if A2 is None:
        nW_A2 = Y.shape[1]
        #A2 = np.eye(Y.shape[1])	#for now this creates A1 and A2
    else:
        nW_A2 = A2.shape[0]


    nW_X1 = X1.shape[1]
    rows_block = nW_A1 * nW_X1

    if 0:#independentX2:
        nW_X2 = 1
    else:
        nW_X2 = X2.shape[1]
    cols_block = nW_A2 * nW_X2

    block = np.zeros((rows_block,cols_block))


    if (R>C) or (A1 is None) or (A2 is None):
        for c in range(C):
            X1D = X1 * D[:,c:c+1]
            X1X2 = X1D.T.dot(X2)
            if (A1 is None) and (A2 is None):
                block[c*X1.shape[1]:(c+1)*X1.shape[1], c*X2.shape[1]:(c+1)*X2.shape[1]] += X1X2
            elif (A1 is None):
                block[c*X1.shape[1]:(c+1)*X1.shape[1],:] += np.kron(A2[:,c:c+1].T,X1X2)
            elif (A2 is None):
                block[:,c*X2.shape[1]:(c+1)*X2.shape[1]] += np.kron(A1[:,c:c+1],X1X2)
            else:
                A1A2 = np.outer(A1[:,c],A2[:,c])
                block += np.kron(A1A2,X1X2)
    else:
        for r in range(R):
            A1D = A1 * D[r:r+1,:]
            A1A2 = A1D.dot(A2.T)
            X1X2 = X1[r,:][:,np.newaxis].dot(X2[r,:][np.newaxis,:])
            block += np.kron(A1A2,X1X2)

    return block

class mean(cObject):

    def __init__(self,Y, identity_trick=False):
        """ init data term """
        self.Y = Y
        self.identity_trick=identity_trick
        self.clearFixedEffect()

    #########################################
    # Properties
    #########################################

    @property
    def A(self):
        return self._A

    @property
    def B(self):
        return self._B

    @property
    def F(self):
        return self._F

    @property
    def A_identity(self):
        return self._A_identity

    @property
    def REML_term(self):
        return self._REML_term

    @property
    def Y(self):
        return self._Y

    @property
    def N(self):
        return self._N

    @property
    def P(self):
        return self._P

    @property
    def n_fixed_effs(self):
        return self._n_fixed_effs

    @property
    def n_terms(self):
        return self._n_terms

    @property
    def Lr(self):
        return self._Lr

    @property
    def Lc(self):
        return self._Lc

    @property
    def d(self):
        return self._d

    @property
    def D(self):
        return np.reshape(self.d,(self.N,self.P), order='F')

    @property
    def LRLdiag(self):
        return self._LRLdiag

    @property
    def LCL(self):
        return self._LCL

    #########################################
    # Setters
    #########################################
    def use_identity_trick(self,identity_trick=True):
        self.identity_trick=identity_trick
        self.clear_cache('Fstar','Astar','Xstar','Xhat',
                         'Areml','Areml_eigh','Areml_chol','Areml_inv','beta_hat','B_hat',
                         'LRLdiag_Xhat_tens','Areml_grad',
                         'beta_grad','Xstar_beta_grad','Zstar','DLZ')


    def clearFixedEffect(self):
        """ erase all fixed effects """
        self._A = []
        self._F = []
        self._B = []
        self._A_identity = []
        self._REML_term = []
        self._n_terms = 0
        self._n_fixed_effs = 0
        self._n_fixed_effs_REML = 0
        self.indicator = {'term':np.array([]),
                            'row':np.array([]),
                            'col':np.array([])}
        self.clear_cache('Fstar','Astar','Xstar','Xhat',
                         'Areml','Areml_eigh','Areml_chol','Areml_inv','beta_hat','B_hat',
                         'LRLdiag_Xhat_tens','Areml_grad',
                         'beta_grad','Xstar_beta_grad','Zstar','DLZ')

    def addFixedEffect(self,F=None,A=None, REML=True, index=None):
        """
        set sample and trait designs
        F:      NxK sample design
        A:      LxP sample design
        REML:   REML for this term?
        index:  index of which fixed effect to replace. If None, just append.
        """
        if F is None:   F = np.ones((self.N,1))
        if A is None:
            A = np.eye(self.P)
            A_identity = True
        elif (A.shape == (self.P,self.P)) & (A==np.eye(self.P)).all():
            A_identity = True
        else:
            A_identity = False

        assert F.shape[0]==self.N, "F dimension mismatch"
        assert A.shape[1]==self.P, "A dimension mismatch"
        if index is None or index==self.n_terms:
            self.F.append(F)
            self.A.append(A)
            self.A_identity.append(A_identity)
            self.REML_term.append(REML)
            # build B matrix and indicator
            self.B.append(np.zeros((F.shape[1],A.shape[0])))
            self._n_terms+=1
            self._update_indicator(F.shape[1],A.shape[0])
        elif index >self.n_terms:
            raise Exception("index exceeds max index of terms")
        else:
            self._n_fixed_effs-=self.F[index].shape[1]*self.A[index].shape[0]
            if self.REML_term[index]:
                self._n_fixed_effs_REML-=self.F[index].shape[1]*self.A[index].shape[0]
            self.F[index] = F
            self.A[index] = A
            self.A_identity[index] = A_identity
            self.REML_term[index]=REML
            self.B[index] = np.zeros((F.shape[1],A.shape[0]))
            self._rebuild_indicator()

        self._n_fixed_effs+=F.shape[1]*A.shape[0]
        if REML:
            self._n_fixed_effs_REML+=F.shape[1]*A.shape[0]
        self.clear_cache('Fstar','Astar','Xstar','Xhat',
                         'Areml','Areml_eigh','Areml_chol','Areml_inv','beta_hat','B_hat',
                         'LRLdiag_Xhat_tens','Areml_grad',
                         'beta_grad','Xstar_beta_grad','Zstar','DLZ')

    def removeFixedEffect(self, index=None):
        """
        set sample and trait designs
        F:      NxK sample design
        A:      LxP sample design
        REML:   REML for this term?
        index:  index of which fixed effect to replace. If None, remove last term.
        """
        if self._n_terms==0:
            pass
        if index is None or index==(self._n_terms-1):

            self._n_terms-=1
            F = self._F.pop() #= self.F[:-1]
            A = self._A.pop() #= self.A[:-1]
            self._A_identity.pop() #= self.A_identity[:-1]
            REML_term = self._REML_term.pop()# = self.REML_term[:-1]
            self._B.pop()# = self.B[:-1]
            self._n_fixed_effs-=F.shape[1]*A.shape[0]
            if REML_term:
                self._n_fixed_effs_REML-=F.shape[1]*A.shape[0]

            pass
        elif index >= self.n_terms:
            raise Exception("index exceeds max index of terms")
        else:
            raise NotImplementedError("currently only last term can be removed")
            pass
        self._rebuild_indicator()
        self.clear_cache('Fstar','Astar','Xstar','Xhat',
                         'Areml','Areml_eigh','Areml_chol','Areml_inv','beta_hat','B_hat',
                         'LRLdiag_Xhat_tens','Areml_grad',
                         'beta_grad','Xstar_beta_grad','Zstar','DLZ')

    @Y.setter
    def Y(self,value):
        """ set phenotype """
        self._N,self._P = value.shape
        self._Y = value
        self.clear_cache('Ystar1','Ystar','Yhat','LRLdiag_Yhat',
                         'beta_grad','Xstar_beta_grad','Zstar','DLZ')

    @Lr.setter
    def Lr(self,value):
        """ set row rotation """
        assert value.shape[0]==self._N, 'dimension mismatch'
        assert value.shape[1]==self._N, 'dimension mismatch'
        self._Lr = value
        self.clear_cache('Fstar','Ystar1','Ystar','Yhat','Xstar','Xhat',
                         'Areml','Areml_eigh','Areml_chol','Areml_inv','beta_hat','B_hat',
                         'LRLdiag_Xhat_tens','LRLdiag_Yhat','Areml_grad',
                         'beta_grad','Xstar_beta_grad',
                         'LRLdiag_Xhat_tens','LRLdiag_Yhat','Areml_grad',
                         'beta_grad','Xstar_beta_grad','Zstar','DLZ')

    @Lc.setter
    def Lc(self,value):
        """ set col rotation """
        assert value.shape[0]==self._P, 'Lc dimension mismatch'
        assert value.shape[1]==self._P, 'Lc dimension mismatch'
        self._Lc = value
        self.clear_cache('Astar','Ystar','Yhat','Xstar','Xhat',
                         'Areml','Areml_eigh','Areml_chol','Areml_inv','beta_hat','B_hat',
                         'LRLdiag_Xhat_tens','LRLdiag_Yhat','Areml_grad',
                         'beta_grad','Xstar_beta_grad','Zstar','DLZ')

    @d.setter
    def d(self,value):
        """ set anisotropic scaling """
        assert value.shape[0]==self._P*self._N, 'd dimension mismatch'
        self._d = value
        self.clear_cache('Yhat','Xhat','Areml','Areml_eigh','Areml_chol','Areml_inv','beta_hat','B_hat',
                         'LRLdiag_Xhat_tens','LRLdiag_Yhat','Areml_grad',
                         'beta_grad','Xstar_beta_grad','Zstar','DLZ')

    @LRLdiag.setter
    def LRLdiag(self,value):
        """ set anisotropic scaling """
        self._LRLdiag = value
        self.clear_cache('LRLdiag_Xhat_tens','LRLdiag_Yhat','Areml_grad',
                         'beta_grad','Xstar_beta_grad')

    @LCL.setter
    def LCL(self,value):
        """ set anisotropic scaling """
        self._LCL = value
        self.clear_cache('Areml_grad','beta_grad','Xstar_beta_grad')

    #########################################
    # Getters (caching)
    #########################################

    @cached
    def Astar(self):
        RV = []
        for term_i in range(self.n_terms):
            RV.append(np.dot(self.A[term_i],self.Lc.T))
        return RV

    @cached
    def Fstar(self):
        RV = []
        for term_i in range(self.n_terms):
            RV.append(np.dot(self.Lr,self.F[term_i]))
        return RV

    def Ystar1(self):
        return np.dot(self.Lr,self.Y)

    @cached
    def Ystar(self):
        return np.dot(self.Ystar1(),self.Lc.T)

    @cached
    def Yhat(self):
        return self.D*self.Ystar()

    @cached
    def Xstar(self):
        RV = np.zeros((self.N*self.P,self.n_fixed_effs))
        ip = 0
        for i in range(self.n_terms):
            Ki = self.A[i].shape[0]*self.F[i].shape[1]
            RV[:,ip:ip+Ki] = np.kron(self.Astar()[i].T,self.Fstar()[i])
            ip += Ki
        return RV

    def var_total(self):
        return (self.Yhat()*self.Ystar()).sum()


    def var_explained(self):
        XKY = self.compute_XKY(M=self.Yhat())
        beta_hat = self.Areml_solve(XKY)
        return (XKY*beta_hat).sum(), beta_hat


    @cached
    def Xhat(self):
        RV = self.d[:,np.newaxis]*self.Xstar()
        return RV

    @cached
    def Areml(self):
        #A1 = self.XstarT_dot(self.Xhat())
        A2 =  self.compute_XKX()
        return A2

    @cached
    def Areml_chol(self):
        return LA.cholesky(self.Areml()).T

    @cached
    def Areml_REML_chol(self):
        return LA.cholesky(self.Areml()).T

    @cached
    def Areml_inv(self):
        return LA.cho_solve((self.Areml_chol(),True),np.eye(self.n_fixed_effs))

    #caching bug:
    #@cached
    def beta_hat(self):
        XKY = self.compute_XKY(M=self.Yhat())
        beta_hat = self.Areml_solve(XKY)
        return beta_hat


    @cached
    def B_hat(self):
        RV = []
        ip = 0
        for term_i in range(self.n_terms):
            RV.append(np.reshape(self.beta_hat()[ip:ip+self.B[term_i].size],self.B[term_i].shape, order='F'))
            ip += self.B[term_i].size
        return RV

    @cached
    def LRLdiag_Xhat_tens(self):
        RV  = np.reshape(self.Xhat(),(self.N,self.P,self.n_fixed_effs),order='F').copy()
        RV *= self.LRLdiag[:,np.newaxis,np.newaxis]
        return RV

    @cached
    def LRLdiag_Yhat(self):
        return self.LRLdiag[:,np.newaxis]*self.Yhat()

    @cached
    def Areml_grad(self):
        RV = np.einsum('jpk,lp->jlk',self.LRLdiag_Xhat_tens(),self.LCL)
        RV = RV.reshape((self.N*self.P,self.n_fixed_effs),order='F')
        RV*= self.d[:,np.newaxis]
        RV = -self.XstarT_dot(RV)
        return RV

    @cached
    def beta_grad(self):
        RV  = np.reshape(np.dot(self.LRLdiag_Yhat(),self.LCL.T),(self.N*self.P),order='F')
        RV *= self.d
        RV  = self.XstarT_dot(RV)
        RV += np.dot(self.Areml_grad(),self.beta_hat())
        RV  = -np.dot(self.Areml_inv(),RV)
        return RV

    @cached
    def Xstar_beta_grad(self):
        RV = np.zeros((self.N,self.P))
        ip = 0
        for term_i in range(self.n_terms):
            _Bgrad = np.reshape(self.beta_grad()[ip:ip+self.B[term_i].size],self.B[term_i].shape, order='F')
            RV+=np.dot(self.Fstar()[term_i],np.dot(_Bgrad,self.Astar()[term_i]))
            ip += self.B[term_i].size
        return RV


    @cached
    def Zstar(self):
        """ predict the value of the fixed effect """
        RV = self.Ystar().copy()
        for term_i in range(self.n_terms):
            if self.identity_trick and self.A_identity[term_i]:
                RV-=np.dot(self.Fstar()[term_i],self.B_hat()[term_i])
            else:
                RV-=np.dot(self.Fstar()[term_i],np.dot(self.B_hat()[term_i],self.Astar()[term_i]))
        self.clear_cache('DLZ')
        return RV

    @cached
    def Areml_eigh(self):
        """compute the eigenvalue decomposition of Astar"""
        s,U = LA.eigh(self.Areml(),lower=True)
        i_pos = (s>1e-10)
        s = s[i_pos]
        U = U[:,i_pos]
        return s,U

    @cached
    def DLZ(self):
        return self.Zstar()*np.reshape(self.D,(self.N,self.P), order='F')

    ###############################################
    # Other getters with no caching, should not they have caching somehow?
    ###############################################

    def Areml_solve(self, b):
        try:
            res = LA.cho_solve((self.Areml_chol(),True),b)
        except LA.LinAlgError:

            s,U = self.Areml_eigh()
            res = U.T.dot(b)
            res /= s[:,np.newaxis]
            res = U.dot(res)

        return res


    def compute_XKY(self, M=None):
        if M is None:
            M = self.Yhat()
        assert M.shape==(self.N,self.P)
        XKY = np.zeros((self.n_fixed_effs))
        n_weights = 0
        for term in range(self.n_terms):
            if self.identity_trick and self.A_identity[term]:
                XKY_block = compute_XYA(DY=M, X=self.Fstar()[term], A=None)
            else:
                XKY_block = compute_XYA(DY=M, X=self.Fstar()[term], A=self.Astar()[term])
            XKY[n_weights:n_weights + self.A[term].shape[0] * self.F[term].shape[1]] = XKY_block.ravel(order='F')
            n_weights += self.A[term].shape[0] * self.F[term].shape[1]
        return XKY

    def compute_XKX(self):
        #n_weights1 = 0
        #
        #for term1 in xrange(self.n_terms):
        #    n_weights1+=self.Astar()[term1].shape[0] * self.Fstar()[term1].shape[1]
        #cov_beta = np.zeros((n_weights1,n_weights1))
        cov_beta = np.zeros((self.n_fixed_effs,self.n_fixed_effs))
        n_weights1 = 0
        for term1 in range(self.n_terms):
            if self.identity_trick and self.A_identity[term1]:
                A_term1 = None
            else:
                A_term1 = self.Astar()[term1]
            n_weights2 = n_weights1
            for term2 in range(term1,self.n_terms):
                if self.identity_trick and self.A_identity[term2]:
                    A_term2 = None
                else:
                    A_term2 = self.Astar()[term2]
                block = compute_X1KX2(Y=self.Ystar(), D=self.D, X1=self.Fstar()[term1], X2=self.Fstar()[term2], A1=A_term1, A2=A_term2)
                cov_beta[n_weights1:n_weights1 + self.A[term1].shape[0] * self.F[term1].shape[1], n_weights2:n_weights2 + self.A[term2].shape[0] * self.F[term2].shape[1]] = block
                if term1!=term2:
                    cov_beta[n_weights2:n_weights2 + self.A[term2].shape[0] * self.F[term2].shape[1], n_weights1:n_weights1 + self.A[term1].shape[0] * self.F[term1].shape[1]] = block.T

                n_weights2+=self.A[term2].shape[0] * self.F[term2].shape[1]

            n_weights1+=self.A[term1].shape[0] * self.F[term1].shape[1]
        return cov_beta

    def predict(self):
        """ predict the value of the fixed effect """
        RV = np.zeros((self.N,self.P))
        for term_i in range(self.n_terms):
            RV+=np.dot(self.Fstar()[term_i],np.dot(self.B()[term_i],self.Astar()[term_i]))
        return RV

    def evaluate(self):
        """ predict the value of """
        RV  = -self.predict()
        RV += self.Ystar()
        return RV

    def getGradient(self,j):
        """ get rotated gradient for fixed effect i """
        i = int(self.indicator['term'][j])
        r = int(self.indicator['row'][j])
        c = int(self.indicator['col'][j])
        rv = -np.kron(self.Fstar()[i][:,[r]],self.Astar()[i][[c],:])
        return rv

    def XstarT_dot(self,M):
        """ get dot product of Xhat and M """
        if 0:
            #TODO: implement this properly
            pass
        else:
            RV = np.dot(self.Xstar().T,M)
        return RV

    def getResiduals(self):
        """ regress out fixed effects and results residuals """
        X = np.zeros((self.N*self.P,self.n_fixed_effs))
        ip = 0
        for i in range(self.n_terms):
            Ki = self.A[i].shape[0]*self.F[i].shape[1]
            X[:,ip:ip+Ki] = np.kron(self.A[i].T,self.F[i])
            ip += Ki
        y = np.reshape(self.Y,(self.Y.size,1),order='F')
        RV = regressOut(y,X)
        RV = np.reshape(RV,self.Y.shape,order='F')
        return RV

    #########################################
    # Params manipulation
    #########################################

    def getParams(self):
        """ get params """
        rv = np.array([])
        if self.n_terms>0:
            rv = np.concatenate([np.reshape(self.B[term_i],self.B[term_i].size, order='F') for term_i in range(self.n_terms)])
        return rv

    def setParams(self,params):
        """ set params """
        start = 0
        for i in range(self.n_terms):
            n_effects = self.B[i].size
            self.B[i] = np.reshape(params[start:start+n_effects],self.B[i].shape, order='F')
            start += n_effects

    #########################################
    # Utility functions
    #########################################

    def getDimensions(self):
        """ get phenotype dimensions """
        return self.N,self.P

    def _set_toChange(x):
        """ set variables in list x toChange """
        for key in list(x.keys()):
            self.toChange[key] = True

    def _update_indicator(self,K,L):
        """ update the indicator """
        _update = {'term': self.n_terms*np.ones((K,L)).T.ravel(),
                    'row': np.kron(np.arange(K)[:,np.newaxis],np.ones((1,L))).T.ravel(),
                    'col': np.kron(np.ones((K,1)),np.arange(L)[np.newaxis,:]).T.ravel()}
        for key in list(_update.keys()):
            self.indicator[key] = np.concatenate([self.indicator[key],_update[key]])

    def _rebuild_indicator(self):
        """ update the indicator """
        indicator = {'term':np.array([]),
                     'row':np.array([]),
                     'col':np.array([])}

        for term in range(self.n_terms):
            L = self.A[term].shape[0]
            K = self.F[term].shape[1]
            _update = {'term': (term+1)*np.ones((K,L)).T.ravel(),
                    'row': np.kron(np.arange(K)[:,np.newaxis],np.ones((1,L))).T.ravel(),
                    'col': np.kron(np.ones((K,1)),np.arange(L)[np.newaxis,:]).T.ravel()}
            for key in list(_update.keys()):
                indicator[key] = np.concatenate([indicator[key],_update[key]])
        self.indicator = indicator



"""@package fdr
FDR estimation using Benjamini Hochberg and Stories method
"""

import numpy as np
import scipy as sp
import sys, pickle, pdb
import scipy.stats as st
import scipy.interpolate
import logging as LG


def qvalues1(PV,m=None,pi=1.0):
    """estimate q vlaues from a list of Pvalues
    this algorihm is taken from Storey, significance testing for genomic ...
    m: number of tests, (if not len(PV)), pi: fraction of expected true null (1.0 is a conservative estimate)
    @param PV: pvalues
    @param m:  total number of tests if PV is not the entire array.
    @param pi: fraction of true null
    """

    S = PV.shape
    PV = PV.flatten()
    if m is None:
        m = len(PV) * 1.0
    else:
        m*=1.0
    lPV = len(PV)

    #1. sort pvalues
    PV = PV.squeeze()
    IPV = PV.argsort()
    PV  = PV[IPV]

    #2. estimate lambda
    if pi is None:
        lrange = sp.linspace(0.05,0.95,max(lPV/100.0,10))
        pil    = sp.double((PV[:,sp.newaxis]>lrange).sum(axis=0))/lPV
        pilr   = pil/(1.0-lrange)
        #ok, I think for SNPs this is pretty useless, pi is close to 1!
        pi =1.0
        #if there is something useful in there use the something close to 1
        if pilr[-1]<1.0:
            pi = pilr[-1]

    #3. initialise q values
    QV_ = pi * m/lPV* PV
    QV_[-1] = min(QV_[-1],1.0)
    #4. update estimate
    for i in range(lPV-2,-1,-1):
        QV_[i] = min(pi*m*PV[i]/(i+1.0),QV_[i+1])
    #5. invert sorting
    QV = sp.zeros_like(PV)
    QV[IPV] = QV_

    QV = QV.reshape(S)
    return QV

def qvalues(pv, m = None, return_pi0 = False, lowmem = False, pi0 = None, fix_lambda = None):

    original_shape = pv.shape

    assert(pv.min() >= 0 and pv.max() <= 1)

    pv = pv.ravel() # flattens the array in place, more efficient than flatten()

    if m == None:
        m = float(len(pv))
    else:
        # the user has supplied an m, let's use it
        m *= 1.0

    # if the number of hypotheses is small, just set pi0 to 1
    if len(pv) < 100:
        pi0 = 1.0
    elif pi0 != None:
        pi0 = pi0
    else:
        # evaluate pi0 for different lambdas
        pi0 = []
        lam = sp.arange(0, 0.90, 0.01)
        counts = sp.array([(pv > i).sum() for i in sp.arange(0, 0.9, 0.01)])

        if fix_lambda != None:
            interv_count = (pv > fix_lambda - 0.01).sum()
            uniform_sim = sp.array([(pv > fix_lambda-0.01).sum()*(i+1) for i in sp.arange(0, len(sp.arange(0, 0.90, 0.01)))][::-1])
            counts += uniform_sim

        for l in range(len(lam)):
            pi0.append(counts[l]/(m*(1-lam[l])))

        pi0 = sp.array(pi0)

        # fit natural cubic spline
        tck = sp.interpolate.splrep(lam, pi0, k = 3)
        pi0 = sp.interpolate.splev(lam[-1], tck)
        if pi0 > 1:
            LG.warning("got pi0 > 1 (%.3f) while estimating qvalues, setting it to 1" % pi0)
            pi0 = 1.0

        assert(pi0 >= 0 and pi0 <= 1), "%f" % pi0


    if lowmem:
        # low memory version, only uses 1 pv and 1 qv matrices
        qv = sp.zeros((len(pv),))
        last_pv = pv.argmax()
        qv[last_pv] = (pi0*pv[last_pv]*m)/float(m)
        pv[last_pv] = -sp.inf
        prev_qv = last_pv
        for i in range(int(len(pv))-2, -1, -1):
            cur_max = pv.argmax()
            qv_i = (pi0*m*pv[cur_max]/float(i+1))
            pv[cur_max] = -sp.inf
            qv_i1 = prev_qv
            qv[cur_max] = min(qv_i, qv_i1)
            prev_qv = qv[cur_max]

    else:
        p_ordered = sp.argsort(pv)
        pv = pv[p_ordered]
        # estimate qvalues
        # qv = pi0*m*pv/(sp.arange(len(pv))+1.0)

        # for i in xrange(int(len(qv))-2, 0, -1):
        #     qv[i] = min([qv[i], qv[i+1]])


    qv = pi0 * m/len(pv) * pv
    qv[-1] = min(qv[-1],1.0)

    for i in range(len(pv)-2, -1, -1):
        qv[i] = min(pi0*m*pv[i]/(i+1.0), qv[i+1])



    # reorder qvalues
    qv_temp = qv.copy()
    qv = sp.zeros_like(qv)
    qv[p_ordered] = qv_temp



    # reshape qvalues
    qv = qv.reshape(original_shape)

    if return_pi0:
        return qv, pi0
    else:
        return qv


def estimate_lambda(pv):
    """estimate lambda form a set of PV"""
    LOD2 = sp.median(st.chi2.isf(pv,1))
    L = (LOD2/0.456)
    return (L)

def LOD2PV(lods):
    PV = (st.chi2.sf(2*lods, 1))
    return PV



import sys
import limix
import scipy as sp
import numpy as np
import numpy.linalg as nla
import scipy.linalg as la
from limix.mtSet.iset_full import ISet_Full
from limix.mtSet.iset_strat import ISet_Strat
import pandas as pd
from limix.mtSet.core.iset_utils import calc_emp_pv_eff
import pdb

def fit_iSet(Y, U_R=None, S_R=None, covs=None, Xr=None, n_perms=0, Ie=None,
             strat=False, verbose=True):
    """
    Args:
        Y:          [N, P] phenotype matrix
        S_R:        N vector of eigenvalues of R
        U_R:        [N, N] eigenvector matrix of R
        covs:       [N, K] matrix for K covariates
        Xr:         [N, S] genotype data of the set component
        n_perms:    number of permutations to consider
        Ie:         N boolean context indicator
        strat:      if True, the implementation with stratified designs is considered
    """
    factr=1e7 # remove?
    if strat:
        assert Ie is not None, 'Ie must be specified for stratification analyses'
        assert Y.shape[1]==1, 'Y must be Nx1 for stratification analysis'
    else:
        assert covs==None, 'Covariates are not supported for analysis of fully observed phenotypes'

    if verbose:     print('fittng iSet')
    if strat:
        mtSetGxE = ISet_Strat(Y, Ie, Xr, covs=covs)
        RV = {}
        RV['null'] = mtSetGxE.fitNull()
        RV['rank2'] = mtSetGxE.fitFullRank()
        RV['rank1'] = mtSetGxE.fitLowRank()
        RV['block'] = mtSetGxE.fitBlock()
        RV['var'] = mtSetGxE.getVC()
    else:
        mtSetGxE = ISet_Full(Y=Y, S_R=S_R, U_R=U_R, Xr=Xr, factr=factr)
        RV = {}
        RV['null'] = mtSetGxE.fitNull()
        RV['rank2'] = mtSetGxE.fitFullRank()
        RV['rank1'] = mtSetGxE.fitLowRank()
        LLR = RV['rank1']['NLLAlt'] - RV['rank2']['NLLAlt']
        if LLR<-1e-6:
            RV['rank2'] = mtSetGxE.fitFullRank(init_method='lr')
        try:
            RV['block'] = mtSetGxE.fitBlock()
        except:
            try:
                RV['block'] = mtSetGxE.fitBlock(init_method='null')
            except:
                RV['block'] = mtSetGxE.fitBlock(init_method='null_no_opt')
        RV['var'] = mtSetGxE.getVC()

    if n_perms>0:
        RVperm = {}
        nulls = ['null', 'block', 'rank1']
        tests = ['mtSet', 'iSet', 'iSet-het']
        for test in tests:
            RVperm[test+' LLR0'] = sp.zeros(n_perms)
        for seed_i in range(n_perms):
            if verbose:     print(('permutation %d / %d' % (seed_i, n_perms)))
            for it, test in enumerate(tests):
                if test=='mtSet':
                    idxs = sp.random.permutation(Xr.shape[0])
                    _Xr = Xr[idxs, :]
                    df0 = fit_iSet(Y, U_R=U_R, S_R=S_R, covs=covs, Xr=_Xr, n_perms=0, Ie=Ie, strat=strat, verbose=False)
                else:
                    Y0 = mtSetGxE._sim_from(set_covar=nulls[it])
                    Y0 -= Y0.mean(0)
                    df0 = fit_iSet(Y0, U_R=U_R, S_R=S_R, covs=covs, Xr=Xr, n_perms=0, Ie=Ie, strat=strat, verbose=False)
                RVperm[test+' LLR0'][seed_i]  = df0[test+' LLR'][0]

    # output
    LLR_mtSet = RV['null']['NLL']-RV['rank2']['NLL']
    LLR_iSet = RV['block']['NLL']-RV['rank2']['NLL']
    LLR_iSet_het = RV['rank1']['NLL']-RV['rank2']['NLL']

    if strat:   var_keys = ['var_r_full', 'var_c', 'var_n']
    else:       var_keys = ['var_r_full', 'var_g', 'var_n']
    varT = sp.sum([RV['var'][key] for key in var_keys])
    var_pers = RV['var']['var_r_block'] / varT
    var_resc = (RV['var']['var_r_rank1'] - RV['var']['var_r_block']) / varT
    var_het = (RV['var']['var_r_full'] - RV['var']['var_r_rank1']) / varT
    conv = RV['null']['conv']
    conv*= RV['block']['conv']
    conv*= RV['rank1']['conv']
    conv*= RV['rank2']['conv']

    M = sp.array([LLR_mtSet, LLR_iSet, LLR_iSet_het, var_pers, var_resc, var_het, conv]).T
    columns = ['mtSet LLR', 'iSet LLR', 'iSet-het LLR',
                'Persistent Var', 'Rescaling-GxC Var', 'Heterogeneity-GxC var', 'Converged']
    df = pd.DataFrame(M, columns=columns)

    if n_perms>0:
        return df, pd.DataFrame(RVperm)
    return df

if __name__=='__main__':

    N = 200
    C = 2
    S = 5
    K = 30
    Y = sp.randn(N,C)
    W = sp.randn(N,K)
    W-= W.mean(0)
    W/= W.std(0)
    R = sp.dot(W, W.T)
    R/= R.diagonal().mean(0)
    R+= 1e-4*sp.eye(N)
    S_R, U_R = nla.eigh(R)
    covs = U_R[:,-10:]
    Ie = sp.rand(200)<0.5

    df = pd.DataFrame()
    df0 = pd.DataFrame()
    n_regions = 5
    for i in range(n_regions):
        print(('.. analyzing region %d' % i))

        X = 1.*(sp.rand(N,S)<0.2)
        X-= X.mean(0)
        X/= X.std(0)
        X/= sp.sqrt(X.shape[1])
        #_df, _df0 = fit_iSet(Y, U_R, S_R, X, n_perms=10)
        _df, _df0 = fit_iSet(Y[:,[0]], Xr=X, covs=covs, n_perms=10, Ie=Ie, strat=True)
        df  = df.append(_df)
        df0 = df0.append(_df0)

    for test in ['mtSet', 'iSet', 'iSet-het']:
        df[test+' pv'] = calc_emp_pv_eff(df[test+' LLR'].values, df0[test+' LLR0'].values)




import unittest
import scipy as sp
import numpy as np
import sys
from limix.core.covar.dirIndirCov import DirIndirCov
from limix.utils.check_grad import mcheck_grad

class TestDirIndirCov(unittest.TestCase):
    def setUp(self):
        sp.random.seed(1)

        # generate data
        n = 10
        f = 2
        X  = 1.*(sp.rand(n,f)<0.2)
        X -= X.mean(0); X /= X.std(0)
        kinship  = sp.dot(X,X.T)
        kinship /= kinship.diagonal().mean()
        design = sp.zeros((n,n))
        for i in range(n//2):
            design[2*i,2*i+1] = 1
            design[2*i+1,2*i] = 1

        # define covariance
        self.C = DirIndirCov(kinship,design)
        self.C.setRandomParams()

    def test_grad(self):
        def func(x, i):
            self.C.setParams(x)
            return self.C.K()

        def grad(x, i):
            self.C.setParams(x)
            return self.C.K_grad_i(i)

        x0 = self.C.getParams()
        err = mcheck_grad(func, grad, x0)

        np.testing.assert_almost_equal(err, 0., decimal=6)

if __name__ == '__main__':
    unittest.main()



# Copyright(c) 2014, The LIMIX developers (Christoph Lippert, Paolo Francesco Casale, Oliver Stegle)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tables
import numpy as np
import time, os
import pandas as pd
from datetime import datetime

def _depth_first_hdf5(dictionary, outfile, root=None, filters=None):
    """
    performs a depth first search on a dict object and creates a HDF5 outfile 
    structure that mirrors the dictionary.
    Supported types for leaf nodes are numpy.ndarray, scalars, regular lists, and pandas.DataFrame
    
    Args:
        dictionary:     dict
        outfile:        pytables HDF5 outfile
        root:           the current root node (default "/")
        filters:        filters for chunked storage
    """
    if root is None:
        root=outfile.root
    for child_key in list(dictionary.keys()):
        if isinstance(dictionary[child_key],dict):
            child_group = outfile.create_group(root, child_key)
            _depth_first_hdf5(dictionary[child_key], outfile=outfile, root=child_group, filters=filters)
        elif np.isscalar(dictionary[child_key]):
            leaf = outfile.create_array(where=root, name=child_key, obj=dictionary[child_key])
        elif isinstance(dictionary[child_key],list):
            leaf = outfile.create_array(where=root, name=child_key, obj=dictionary[child_key])
        elif isinstance(dictionary[child_key],np.ndarray):
            atom = tables.Atom.from_dtype(dictionary[child_key].dtype)
            if filters is None:
                leaf = outfile.create_array(where=root, name=child_key, atom=atom, shape=dictionary[child_key].shape, obj=dictionary[child_key])
            else:
                leaf = outfile.create_array(where=root, name=child_key, atom=atom, shape=dictionary[child_key].shape, obj=dictionary[child_key], filters=filters)
        elif isinstance(dictionary[child_key],pd.core.frame.DataFrame):
            #raise NotImplementedError("to columscome: DataFrames to HDF")
            child_group = outfile.create_group(root, child_key) #The DataFrame is stored in a group of its own

            index = np.array(dictionary[child_key].index,dtype=type(dictionary[child_key].index[0]))
            leaf = outfile.create_array(where=child_group, name="index", obj=index) #Store the index
            
            columns = np.array(dictionary[child_key].columns,dtype=type(dictionary[child_key].columns[0]))
            leaf = outfile.create_array(where=child_group, name="columns", obj=columns) #store the columns
            
            values = dictionary[child_key].values
            leaf = outfile.create_array(where=child_group, name="values", obj=columns)  #store the values
        else:
            raise IOError("unsupported IO type in output dictionary: "+str(type(dictionary)))

def _depth_first_text(dictionary, outdir=".", delimiter=" ",float_format="%.6e"):
    """
    performs a depth first search on a dict object and creates a directory 
    structure that mirrors the dictionaries. The leafs of the dictionaries are text files
    Supported types for leaf nodes are numpy.ndarray, scalars, regular lists and pandas.DataFrame
    
    Args:
        dictionary:     dict
        outdir:         directory to write in
        delimiter:      delimiter for root node text files
        float_format:   default format for floating point outputs.   
    """
    
    for child_key in list(dictionary.keys()):
        if isinstance(dictionary[child_key],dict):
            child_path = os.path.join(outdir, child_key)
            _depth_first_text(dictionary[child_key], outdir=child_path,delimiter=delimiter,float_format=float_format)
        else:
            if not os.path.exists(outdir):
                os.makedirs(outdir)
            filename = os.path.join(outdir, child_key)+".txt"
            if np.isscalar(dictionary[child_key]):
                file = open(filename, "w")
                if type(dictionary[child_key])==float:
                    outstr=float_format % dictionary[child_key]
                else:
                    outstr=str(dictionary[child_key])
                file.write(outstr)
                file.close()
            elif isinstance(dictionary[child_key],list):
                outarray=np.array(dictionary[child_key])
                _write_txt_array(filename,array=outarray,delimiter=delimiter,float_format=float_format)
            elif isinstance(dictionary[child_key],np.ndarray):
                _write_txt_array(filename,array=dictionary[child_key],delimiter=delimiter,float_format=float_format)
            elif isinstance(dictionary[child_key],pd.core.frame.DataFrame):
                dictionary[child_key].to_csv(filename,sep=delimiter,index=True,header=True,na_rep="nan",float_format=float_format)
            else:
                raise IOError("unsupported IO type in output dictionary: "+str(type(dictionary)))

def _write_txt_array(filename, array, delimiter=" ",float_format="%.6f"):
    """
    stores an array with unknown dtype to a text file.
    
    Args:
        filename:       name of the text file
        array:          array to be saved in the text file
        delimiter:      delimiter seperating values in the text file
        float_format:   formating string for floating point values
    """
    if array.dtype==np.float:
        np.savetxt(filename,array,fmt=float_format,delimiter=delimiter)
    elif array.dtype==np.integer:
        np.savetxt(filename,array,fmt="%i",delimiter=delimiter)
    elif array.dtype.type==np.string_:
        np.savetxt(filename,array,fmt="%s",delimiter=delimiter)
    else:
        raise IOError("unsupported format")

class output_writer(object):
    """writes an output dictionary to disk"""
    
    def __init__(self, output_dictionary,timestamp=None):
        """
        Args:
            output_dictionary:  a dictionary holding either other dictionaries, numpy arrays, or scalars as members
            timestamp:          a timestamp from time.time() If None, the current time is used. 
        """
        self.output_dict = output_dictionary
        self.filters = None
        if timestamp is None:
            self.timestamp = time.time()
        else:
            self.timestamp = timestamp
        #here we could add filters for saving compressed chunked arrays:
        #self.filters = tables.Filters(complib='blosc', complevel=5)

    def get_timestamp(self):
        """
        create a timestamp
        Returns:
            a string 
        """
        return str(datetime.fromtimestamp(self.timestamp))[0:10]+"_"+str(datetime.fromtimestamp(self.timestamp))[11:13]+"-"+str(datetime.fromtimestamp(self.timestamp))[14:16]+"-"+str(datetime.fromtimestamp(self.timestamp))[17:19]

    def write_hdf5(self, filename, timestamp=False):
        """
        Creates a HDF5 file that mirrors the dictionary.
        Supported types for leaf nodes are numpy.ndarray, scalars, and regular lists
    
        Args:
            filename:       name of the HDF5 output file
            timestamp:      Boolean indicator whether to append a timestap to the filename
        """
        if timestamp:
            stamp=self.get_timestamp()
            if len(filename)>4 and filename[-4] == ".":
                filename=filename[0:-4]+"_"+self.get_timestamp()+filename[-4:]
            elif len(filename)>3 and filename[-3] == ".":
                filename=filename[0:-3]+"_"+self.get_timestamp()+filename[-3:]
            else:
                filename=filename+self.get_timestamp()
        outfile = tables.open_file(filename, mode = "w", title = "Output")
        _depth_first_hdf5(self.output_dict,outfile=outfile,root=outfile.root,filters=self.filters)
        outfile.close()
        pass
    def write_txt(self, outdir=".", delimiter=" ",float_format="%.6f", timestamp=False):
        """
        Creates a directory structure that mirrors the dictionary.
        The leaf nodes are text files
        Supported types for leaf nodes are numpy.ndarray, scalars, and regular lists
    
        Args:
            outdir:         name of the HDF5 output file
            delimiter:      delimiter for values in text files
            float_format:   formating string for floating point values
            timestamp:      Boolean indicator whether to append a timestap to the filename
        """
        if timestamp:
            stamp=self.get_timestamp()
            if outdir.endswith(".") or outdir=="" or outdir.endswith(".."):
                outdir=os.path.join(outdir,self.get_timestamp())
            else:
                outdir=outdir+"_"+self.get_timestamp()
        _depth_first_text(dictionary=self.output_dict, outdir=outdir, delimiter=delimiter,float_format=float_format)
        pass


if __name__ == "__main__":
    print(("last modified: %s" % time.ctime(os.path.getmtime(__file__))))
    di={"a":{"B":np.ones((5,5)),"C":1},"B":["dfd","f"]}
    writer = output_writer(output_dictionary=di)
    writer.write_hdf5("test.h5",timestamp=True)
    writer.write_txt("test",timestamp=True)


# import unittest
# from limix.core.covar import FreeFormCov
# from limix.core.mean import MeanKronSum
# from limix.core.gp import GP2KronSumLR
# from limix.core.gp import GP
# from limix.utils.preprocess import covar_rescale
# from limix.utils.check_grad import mcheck_grad
#
# import numpy as np
# import scipy as sp
# import pdb
#
#
# class TestGPBase(unittest.TestCase):
#     def setUp(self):
#         np.random.seed(1)
#
#         # define phenotype
#         N = 200
#         P = 2
#         Y = sp.randn(N,P)
#         # define fixed effects
#         F = []; A = []
#         F.append(1.*(sp.rand(N,2)<0.5))
#         A.append(sp.eye(P))
#         # define row caoriance
#         f = 10
#         G = 1.*(sp.rand(N, f)<0.2)
#         # define col covariances
#         Cr = FreeFormCov(P)
#         self._Cr = Cr
#         Cn = FreeFormCov(P)
#         Cr.setCovariance(0.5 * sp.cov(Y.T))
#         Cn.setCovariance(0.5 * sp.cov(Y.T))
#         # define gp
#         self.gp = GP2KronSumLR(Y = Y, F = F, A = A, Cn = Cn, G = G)
#
#     def test_grad(self):
#
#         gp = self.gp
#
#         def func(x, i):
#             params = gp.getParams()
#             params['covar'] = x
#             gp.setParams(params)
#             return gp.LML()
#
#         def grad(x, i):
#             params = gp.getParams()
#             params['covar'] = x
#             gp.setParams(params)
#             grad = gp.LML_grad()
#             return grad['covar'][i]
#
#         x0 = gp.getParams()['covar']
#         err = mcheck_grad(func, grad, x0)
#         np.testing.assert_almost_equal(err, 0., decimal=4)
#
#     def test_grad_activation(self):
#
#         gp = self.gp
#
#         # gp.covar.act_Cr = False
#         self._Cr.act_K = False
#
#         def func(x, i):
#             params = gp.getParams()
#             params['covar'] = x
#             gp.setParams(params)
#             return gp.LML()
#
#         def grad(x, i):
#             params = gp.getParams()
#             params['covar'] = x
#             gp.setParams(params)
#             grad = gp.LML_grad()
#             return grad['covar'][i]
#
#         x0 = gp.getParams()['covar']
#         err = mcheck_grad(func, grad, x0)
#         np.testing.assert_almost_equal(err, 0., decimal=4)
#
#
# if __name__ == "__main__":
#     unittest.main()



#!/usr/bin/env python
"""Doxygen XML to SWIG docstring converter.

Converts Doxygen generated XML files into a file containing docstrings
that can be used by SWIG-1.3.x.  Note that you need to get SWIG
version > 1.3.23 or use Robin Dunn's docstring patch to be able to use
the resulting output.

Usage:

  doxy2swig.py input.xml output.i

input.xml is your doxygen generated XML file and output.i is where the
output will be written (the file will be clobbered).

"""

# This code is implemented using Mark Pilgrim's code as a guideline:
#   http://www.faqs.org/docs/diveintopython/kgp_divein.html
#
# Author: Prabhu Ramachandran
# License: BSD style


from xml.dom import minidom
import re
import textwrap
import sys
import types
import os.path


def my_open_read(source):
    if hasattr(source, "read"):
        return source
    else:
        return open(source)

def my_open_write(dest, mode='w'):
    if hasattr(dest, "write"):
        return dest
    else:
        return open(dest, mode)


class Doxy2SWIG:
    """Converts Doxygen generated XML files into a file containing
    docstrings that can be used by SWIG-1.3.x that have support for
    feature("docstring").  Once the data is parsed it is stored in
    self.pieces.

    """

    def __init__(self, src):
        """Initialize the instance given a source object (file or
        filename).

        """
        f = my_open_read(src)
        self.my_dir = os.path.dirname(f.name)
        self.xmldoc = minidom.parse(f).documentElement
        f.close()

        self.pieces = []
        self.pieces.append('\n// File: %s\n'%\
                           os.path.basename(f.name))

        self.space_re = re.compile(r'\s+')
        self.lead_spc = re.compile(r'^(%feature\S+\s+\S+\s*?)"\s+(\S)')
        self.multi = 0
        self.ignores = ('inheritancegraph', 'param', 'listofallmembers',
                        'innerclass', 'name', 'declname', 'incdepgraph',
                        'invincdepgraph', 'programlisting', 'type',
                        'references', 'referencedby', 'location',
                        'collaborationgraph', 'reimplements',
                        'reimplementedby', 'derivedcompoundref',
                        'basecompoundref')
        #self.generics = []

    def generate(self):
        """Parses the file set in the initialization.  The resulting
        data is stored in `self.pieces`.

        """
        self.parse(self.xmldoc)

    def parse(self, node):
        """Parse a given node.  This function in turn calls the
        `parse_<nodeType>` functions which handle the respective
        nodes.

        """
        pm = getattr(self, "parse_%s"%node.__class__.__name__)
        pm(node)

    def parse_Document(self, node):
        self.parse(node.documentElement)

    def parse_Text(self, node):
        txt = node.data
        txt = txt.replace('\\', r'\\\\')
        txt = txt.replace('"', r'\"')
        # ignore pure whitespace
        m = self.space_re.match(txt)
        if m and len(m.group()) == len(txt):
            pass
        else:
            self.add_text(textwrap.fill(txt))

    def parse_Element(self, node):
        """Parse an `ELEMENT_NODE`.  This calls specific
        `do_<tagName>` handers for different elements.  If no handler
        is available the `generic_parse` method is called.  All
        tagNames specified in `self.ignores` are simply ignored.

        """
        name = node.tagName
        ignores = self.ignores
        if name in ignores:
            return
        attr = "do_%s" % name
        if hasattr(self, attr):
            handlerMethod = getattr(self, attr)
            handlerMethod(node)
        else:
            self.generic_parse(node)
            #if name not in self.generics: self.generics.append(name)

    def add_text(self, value):
        """Adds text corresponding to `value` into `self.pieces`."""
        if type(value) in (list, tuple):
            self.pieces.extend(value)
        else:
            self.pieces.append(value)

    def get_specific_nodes(self, node, names):
        """Given a node and a sequence of strings in `names`, return a
        dictionary containing the names as keys and child
        `ELEMENT_NODEs`, that have a `tagName` equal to the name.

        """
        nodes = [(x.tagName, x) for x in node.childNodes \
                 if x.nodeType == x.ELEMENT_NODE and \
                 x.tagName in names]
        return dict(nodes)

    def generic_parse(self, node, pad=0):
        """A Generic parser for arbitrary tags in a node.

        Parameters:

         - node:  A node in the DOM.
         - pad: `int` (default: 0)

           If 0 the node data is not padded with newlines.  If 1 it
           appends a newline after parsing the childNodes.  If 2 it
           pads before and after the nodes are processed.  Defaults to
           0.

        """
        npiece = 0
        if pad:
            npiece = len(self.pieces)
            if pad == 2:
                self.add_text('\n')
        for n in node.childNodes:
            self.parse(n)
        if pad:
            if len(self.pieces) > npiece:
                self.add_text('\n')

    def space_parse(self, node):
        self.add_text(' ')
        self.generic_parse(node)

    do_ref = space_parse
    do_emphasis = space_parse
    do_bold = space_parse
    do_computeroutput = space_parse
    do_formula = space_parse

    def do_compoundname(self, node):
        self.add_text('\n\n')
        data = node.firstChild.data
        self.add_text('%%feature("docstring") %s "\n'%data)

    def do_compounddef(self, node):
        kind = node.attributes['kind'].value
        if kind in ('class', 'struct'):
            prot = node.attributes['prot'].value
            if prot != 'public':
                return
            names = ('compoundname', 'briefdescription',
                     'detaileddescription', 'includes')
            first = self.get_specific_nodes(node, names)
            for n in names:
                if n in first:
                    self.parse(first[n])
            self.add_text(['";','\n'])
            for n in node.childNodes:
                if n not in list(first.values()):
                    self.parse(n)
        elif kind in ('file', 'namespace'):
            nodes = node.getElementsByTagName('sectiondef')
            for n in nodes:
                self.parse(n)

    def do_includes(self, node):
        self.add_text('C++ includes: ')
        self.generic_parse(node, pad=1)

    def do_parameterlist(self, node):
        self.add_text(['\n', '\n', 'Parameters:', '\n'])
        self.generic_parse(node, pad=1)

    def do_para(self, node):
        self.add_text('\n')
        self.generic_parse(node, pad=1)

    def do_parametername(self, node):
        self.add_text('\n')
        self.add_text("%s: "%node.firstChild.data)

    def do_parameterdefinition(self, node):
        self.generic_parse(node, pad=1)

    def do_detaileddescription(self, node):
        self.generic_parse(node, pad=1)

    def do_briefdescription(self, node):
        self.generic_parse(node, pad=1)

    def do_memberdef(self, node):
        prot = node.attributes['prot'].value
        id = node.attributes['id'].value
        kind = node.attributes['kind'].value
        tmp = node.parentNode.parentNode.parentNode
        compdef = tmp.getElementsByTagName('compounddef')[0]
        cdef_kind = compdef.attributes['kind'].value

        if prot == 'public':
            first = self.get_specific_nodes(node, ('definition', 'name'))
            name = first['name'].firstChild.data
            if name[:8] == 'operator': # Don't handle operators yet.
                return

            defn = first['definition'].firstChild.data
            self.add_text('\n')
            self.add_text('%feature("docstring") ')

            anc = node.parentNode.parentNode
            if cdef_kind in ('file', 'namespace'):
                ns_node = anc.getElementsByTagName('innernamespace')
                if not ns_node and cdef_kind == 'namespace':
                    ns_node = anc.getElementsByTagName('compoundname')
                if ns_node:
                    ns = ns_node[0].firstChild.data
                    self.add_text(' %s::%s "\n%s'%(ns, name, defn))
                else:
                    self.add_text(' %s "\n%s'%(name, defn))
            elif cdef_kind in ('class', 'struct'):
                # Get the full function name.
                anc_node = anc.getElementsByTagName('compoundname')
                cname = anc_node[0].firstChild.data
                self.add_text(' %s::%s "\n%s'%(cname, name, defn))

            for n in node.childNodes:
                if n not in list(first.values()):
                    self.parse(n)
            self.add_text(['";', '\n'])

    def do_definition(self, node):
        data = node.firstChild.data
        self.add_text('%s "\n%s'%(data, data))

    def do_sectiondef(self, node):
        kind = node.attributes['kind'].value
        if kind in ('public-func', 'func'):
            self.generic_parse(node)

    def do_simplesect(self, node):
        kind = node.attributes['kind'].value
        if kind in ('date', 'rcs', 'version'):
            pass
        elif kind == 'warning':
            self.add_text(['\n', 'WARNING: '])
            self.generic_parse(node)
        elif kind == 'see':
            self.add_text('\n')
            self.add_text('See: ')
            self.generic_parse(node)
        else:
            self.generic_parse(node)

    def do_argsstring(self, node):
        self.generic_parse(node, pad=1)

    def do_member(self, node):
        kind = node.attributes['kind'].value
        refid = node.attributes['refid'].value
        if kind == 'function' and refid[:9] == 'namespace':
            self.generic_parse(node)

    def do_doxygenindex(self, node):
        self.multi = 1
        comps = node.getElementsByTagName('compound')
        for c in comps:
            refid = c.attributes['refid'].value
            fname = refid + '.xml'
            if not os.path.exists(fname):
                fname = os.path.join(self.my_dir,  fname)
            print("parsing file: %s"%fname)
            p = Doxy2SWIG(fname)
            p.generate()
            self.pieces.extend(self.clean_pieces(p.pieces))

    def write(self, fname, mode='w'):
        o = my_open_write(fname, mode)
        if self.multi:
            o.write("".join(self.pieces))
        else:
            o.write("".join(self.clean_pieces(self.pieces)))
        o.close()

    def clean_pieces(self, pieces):
        """Cleans the list of strings given as `pieces`.  It replaces
        multiple newlines by a maximum of 2 and returns a new list.
        It also wraps the paragraphs nicely.

        """
        ret = []
        count = 0
        for i in pieces:
            if i == '\n':
                count = count + 1
            else:
                if i == '";':
                    if count:
                        ret.append('\n')
                elif count > 2:
                    ret.append('\n\n')
                elif count:
                    ret.append('\n'*count)
                count = 0
                ret.append(i)

        _data = "".join(ret)
        ret = []
        for i in _data.split('\n\n'):
            if i == 'Parameters:':
                ret.extend(['Parameters:\n-----------', '\n\n'])
            elif i.find('// File:') > -1: # leave comments alone.
                ret.extend([i, '\n'])
            else:
                _tmp = textwrap.fill(i.strip())
                _tmp = self.lead_spc.sub(r'\1"\2', _tmp)
                ret.extend([_tmp, '\n\n'])
        return ret


def main(input, output):
    p = Doxy2SWIG(input)
    p.generate()
    p.write(output)


if __name__ == '__main__':
    if len(sys.argv) != 3:
        print(__doc__)
        sys.exit(1)
    main(sys.argv[1], sys.argv[2])



def getClient():
    from IPython.parallel import Client
    import os
    sys_name = os.uname()[1]
    if sys_name[:3] == 'ebi':
        return Client(profile='lsf')
    else:
        return Client()


def setPaths(projects_path='~/my_research/region_testing'):
    import sys
    import os
    paths = []
    modulePath = os.path.expanduser(projects_path)
    paths.append(modulePath)
    sys.path.extend(paths)
    return sys.path
    
def par_init(tc, **rem_init_parameters):
    print(('starting mixed forest on', len(tc), 'nodes'))
    init_parameters = rem_init_parameters.copy()
    n_trees = [init_parameters['n_estimators'] // len(tc)] * len(tc)
    for i in range(init_parameters['n_estimators'] % len(tc)):
        n_trees[i] += 1
    dview = tc[:]
    dview.block = True
    dview.apply(setPaths)[0]
    dview.execute('from mixed_forest import MixedForest')
    dview.execute('from mixed_forest.parMixedForest import *')
    lbview = tc.load_balanced_view()
    jobs = []
    for n_tree in n_trees:
        init_parameters['n_estimators'] = n_tree
        job = lbview.apply_async(init, *[init_parameters])
        jobs.append(job)
    tc.wait(jobs)
    
def init(init_parameters):
    from mixed_forest import MixedForest
    globals()['forest']= MixedForest.Forest(**init_parameters)
    return True

def par_build_stumps(tc, X, y, kernel, delta):
    dview = tc[:]
    dview.block = True
    jobs = dview.apply(build_stumps, *[X, y, kernel, delta])
    #for job in jobs:
        #print job

def build_stumps(X, y, kernel, delta):
    globals()['forest'].build_stumps(X, y, kernel, delta)
    return True

def par_fixed_effect(tc, X, oob, depth):
    import scipy as SP
    dview = tc[:]
    dview.block = True
    results = dview.apply(fixed_effect, *[X, oob, depth])
    fixed_sum = SP.zeros_like(results[0][0])
    count = SP.zeros_like(results[0][1])
    for res in results:
        fixed_sum += res[0]
        count += res[1]
    return fixed_sum, count

def clearMemory(tc):
    dview = tc[:]
    dview.block=True
    dview.apply(clearall)
    dview.results.clear()
    tc.metadata.clear()
    tc.results.clear()

def clearall():
    """clear all globals"""
    for uniquevar in [var for var in globals().copy() if var[0] != "_" and var != 'clearall']:
        del globals()[uniquevar]
        
def fixed_effect(X, oob, depth):
    return globals()['forest'].__fixed_effect__(X, oob, depth)

def par_further(tc, depth):
    import scipy as SP
    dview = tc[:]
    dview.block = True
    depths = SP.array(dview.apply(further, *[depth]))
    return depths[SP.argmax(depths)]
     
def further(depth):
    globals()['forest'].further(depth)
    return globals()['forest'].depth

def par_update_delta(tc, delta):
    dview = tc[:]
    dview.block = True
    dview.apply(update_delta, *[delta])
    
def update_delta(delta):
    globals()['forest'].delta = delta
    return True

def par_get_variable_scores(tc):
    import scipy as SP
    dview = tc[:]
    dview.block = True
    results = dview.apply(get_variable_scores)
    var_used = SP.zeros_like((results[0])[0])
    log_importance = SP.zeros_like(var_used)
    for result in results:
        var_used +=  result[0]
        log_importance += result[1]
    return var_used, log_importance
        
def get_variable_scores():
    return globals()['forest'].var_used, globals()['forest'].log_importance


# #!/usr/bin/env python
#
# import nlopt # THIS IS NOT A PACKAGE!
# import numpy as np
#
# print(('nlopt version='+nlopt.__version__))
#
# def f(x, grad):
#     F=x[0]
#     L=x[1]
#     E=x[2]
#     I=x[3]
#     D=F*L**3/(3.*E*I)
#     return D
#
# n = 4
# opt = nlopt.opt(nlopt.LN_COBYLA, n)
# opt.set_min_objective(f)
# lb = np.array([40., 50., 30e3, 1.])
# ub = np.array([60., 60., 40e3, 10.])
# x = (lb+ub)/2.
# opt.set_lower_bounds(lb)
# opt.set_upper_bounds(ub)
# opt.set_xtol_rel(1e-3)
# opt.set_ftol_rel(1e-3)
# xopt = opt.optimize(x)
#
# opt_val = opt.last_optimum_value()
# result = opt.last_optimize_result()
# print(('opt_result='+str(result)))
# print(('optimizer='+str(xopt)))
# print(('opt_val='+str(opt_val)))



__author__ = 'damienarnol1'


import numpy as np

from limix.core.covar import SQExpCov
from limix.core.covar import ZKZCov
from limix.core.covar import FixedCov
from limix.core.covar import SumCov
from limix.core.gp import GP
from limix.core.mean import MeanBase
from limix.utils.preprocess import covar_rescaling_factor


def normalise(mat, axis=1):
    mat -= np.reshape(mat.mean(axis=axis), [mat.shape[0], 1])
    mat /= np.reshape(mat.std(axis=axis), [mat.shape[0], 1])
    mat /= np.sqrt(mat.shape[1])
    return mat


def run_individual_model(model, expression_file, position_file, output_directory,
                         permute_positions=False, random_start_point=False):

    rm_diag = True

    if model is not 'full' and model is not 'env':
        raise Exception('model not understood. Please specify a model between full and env')

    # read phenotypes data
    with open(expression_file, 'r') as f:
        prot_tmp = f.readline()
    protein_names = prot_tmp.split(' ')
    protein_names[-1] = protein_names[-1][0:-1]  # removing the newline sign at the end of the last protein
    protein_names = np.reshape(protein_names, [len(protein_names), 1])
    phenotypes = np.loadtxt(expression_file, delimiter=' ', skiprows=1)

    # read position data
    X = np.genfromtxt(position_file, delimiter=',')
    if permute_positions:
        X = X[np.random.permutation(X.shape[0]), :]
    if X.shape[0] != phenotypes.shape[0]:
        raise Exception('cell number inconsistent between position and epression levels ')

    # define output file name
    output_file = output_directory+'/inferred_parameters_' + model
    if permute_positions:
        output_file += '_permuted.txt'
    else:
        output_file += '.txt'

    N_cells = phenotypes.shape[0]

    parameters = np.zeros([phenotypes.shape[1], 6])

    log_lik = np.zeros(phenotypes.shape[1])

    for phen in range(0, phenotypes.shape[1]):

        phenotype = phenotypes[:, phen]
        phenotype -= phenotype.mean()
        phenotype /= phenotype.std()
        phenotype = np.reshape(phenotype, [N_cells, 1])

        phenotypes_tmp = np.delete(phenotypes, phen, axis=1)
        phenotypes_tmp = normalise(phenotypes_tmp)

        Kinship = phenotypes_tmp.dot(phenotypes_tmp.transpose())
        Kinship -= np.linalg.eigvalsh(Kinship).min() * np.eye(N_cells)
        Kinship *= covar_rescaling_factor(Kinship)

        # create different models and print the result including likelihood
        # create all the covariance terms
        direct_cov = FixedCov(Kinship)

        # noise
        noise_cov = FixedCov(np.eye(N_cells))

        # local_noise
        local_noise_cov = SQExpCov(X)
        local_noise_cov.length = 100
        local_noise_cov.act_length = False
        # environment effect
        environment_cov = ZKZCov(X, Kinship, rm_diag)

        # mean term
        mean = MeanBase(phenotype)

        #######################################################################
        # defining model
        #######################################################################
        cov = SumCov(noise_cov, local_noise_cov)
        cov = SumCov(cov, environment_cov)
        if random_start_point:
            environment_cov.length = np.random.uniform(10, 300)
            environment_cov.scale = np.random.uniform(1, 15)

        else:
            environment_cov.length = 200
        # environment_cov.act_length = False

        if model == 'full':
            cov = SumCov(cov, direct_cov)
        else:
            direct_cov.scale = 0

        # define and optimise GP
        gp = GP(covar=cov, mean=mean)

        try:
            gp.optimize()
        except:
            print('optimisation', str(phen), 'failed')
            continue

        log_lik[phen] = gp.LML()


        # rescale each terms to sample variance one
        # direct cov: unnecessary as fixed covariance rescaled before optimisation
        # local noise covariance
        tmp = covar_rescaling_factor(local_noise_cov.K()/local_noise_cov.scale)
        local_noise_cov.scale /= tmp
        # env effect
        tmp = covar_rescaling_factor(environment_cov.K()/environment_cov.scale**2)
        environment_cov.scale = environment_cov.scale**2/tmp

        parameters[phen, :] = [direct_cov.scale,
                               noise_cov.scale,
                               local_noise_cov.scale,
                               local_noise_cov.length,
                               environment_cov.scale,
                               environment_cov.length]

    result_header = 'direct_scale' + ' ' + \
                    'noise_scale' + ' ' + \
                    'local_noise_scale' + ' ' + \
                    'local_noise_length' + ' ' + \
                    'environment_scale' + ' ' + \
                    'environment_length'

    with open(output_file, 'w') as f:
        np.savetxt(f,
                   np.hstack((protein_names, parameters)),
                   delimiter=' ',
                   header=result_header,
                   fmt='%s',
                   comments='')

    log_lik_file = output_file + '_loglik'
    with open(log_lik_file, 'w') as f:
        np.savetxt(f, log_lik)





from .sampler import standardize_design
from .sampler import standardize_cov
from .sampler import TraitSampler
from .sampler import BernoulliTraitSampler
from .sampler import BinomialTraitSampler
from .sampler import binary_effsiz_sampler
from .sampler import normal_effsiz_sampler
from .sampler import static_effsiz_sampler
from .sampler import geometric_ntrials_sampler
